import multiprocessing.pool as mp
import os
from typing import Callable, Iterable, Optional


class Pool:
    """Wrapper class for multiprocessing pool. If n_processes=1 just use map on main
    thread for debugging purposes """

    def __init__(
            self,
            processes: Optional[int]=None,
            initializer: Optional[Callable]=None,
            initargs: Optional[Iterable]=None,
            *args, **kwargs):
        self.initializer = initializer
        self.initargs = initargs or []
        if processes == 1 or os.name == "nt":
            self.pool = None
        else:
            self.pool = mp.Pool(processes, self.initializer, self.initargs, *args, **kwargs)

    def map(self, func, iterable, chunksize=None):
        if self.pool is None:
            self.initializer(*self.initargs)
            return map(func, iterable)
        return self.pool.map(func, iterable, chunksize)

    def imap(self, func, iterable, chunksize=1):
        if self.pool is None:
            self.initializer(*self.initargs)
            return map(func, iterable)
        return self.pool.imap(func, iterable, chunksize)

    def __enter__(self, *args, **kwargs):
        if self.pool is None:
            return self
        return self.pool.__enter__(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        if self.pool is None:
            return None
        return self.pool.__exit__(*args, **kwargs)



from abc import abstractmethod
from typing import Any, Callable, Optional, Sequence, Tuple, Type, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.pipeline.pipelinecomponent import PipelineComponent
from starfish.core.types import Axes, Number, SpotAttributes
from starfish.core.util import click
from starfish.core.util.click.indirectparams import ImageStackParamType


class DetectSpots(PipelineComponent):
    """
    Starfish spot detectors use a variety of means to detect bright spots against
    dark backgrounds. Starfish's spot detectors each have different strengths and weaknesses.

    **Fixed-position spot finders**

    The following spot finders have two modes of operation.

    The first mode is suitable for coded
    experiments where genes are identified by patterns of spots over all rounds and channels of the
    experiment. In this mode, the spot finders identify spots in a single image, which can be either
    a dots auxiliary image, or a maximum intensity projection of the primary images.
    The positions of the maxima are then measured in all other images, and the intensities across
    the complete experiment are stored in an :ref:`IntensityTable`

    The second mode is suitable for assays that detect spots in a single round, such as single
    molecule FISH and RNAscope. This mode simply finds all the spots and concatenates them into a
    long-form IntensityTable. In this mode, the spots are not measured in images that correspond to
    other :code:`(round, channel)` pairs; those positions of the IntensityTable are filled with
    :code:`np.nan`.

    1. The :py:class:`~starfish.spots._detect_spots.blob.BlobDetector` allows the user to pre-filter
    an image using either a Laplacian-of-Gaussians or
    Difference-of-Gaussians (fast approximation to Laplacian-of-Gaussians). These filters are
    applied at with a user-specified variety of Gaussian kernel sizes, and the best-fitting size is
    automatically selected. This allows this filter to detect Gaussian shaped blobs of various
    sizes.

    2. The :py:class:`~starfish.spots._detect_spots.local_max_peak_finder.LocalMaxPeakFinder`
    identifies local maxima using the same machinery as the BlobDetector, except that
    it requires that the user to pre-apply any filters to enhance spots. In exchange, it allows a
    user to automatically select the threshold that separates foreground (spots) from background
    (noise).

    In the future, starfish will combine the functionality of LocalMaxPeakFinder into BlobDetector
    so that a user can detect blobs of multiple sizes *and* automatically find a stable threshold.

    3. The
    :py:class:`~starfish.spots._detect_spots.trackpy_local_max_peak_finder.TrackpyLocalMaxPeakFinder`
    provides an implementation of the `Crocker-Grier <crocker_grier>`_ spot finding
    algorithm. This method optionally preprocesses the image by performing a band pass and a
    threshold. It then locates all peaks of brightness, characterizes the neighborhoods of the peaks
    and takes only those with given total brightness (“mass”). Finally, it refines the positions of
    each peak.

    .. _crocker_grier: https://physics.nyu.edu/grierlab/methods3c/

    **Fuzzy-position spot finders**

    In addition to the spot finders above, we expose single additional spot finder that, while very
    similar in implementation to the BlobFinder, is able to adjust to small local spatial
    perturbations in the centroid of the spot across rounds and channels, such as those that might
    occur from small shifts to the tissue or stage.

    1. The
    :py:class:`~starfish.spots._detect_spots.local_search_blob_detector.LocalSearchBlobDetector`
    is a Gaussian blob detector finds spots in all rounds and channels independently, then, given
    each spot in a user specified "anchor round", selects the closest spot by spatial position in
    all other rounds and aggregates those into codes which can subsequently be decoded. This Spot
    detector is only applicable to experiments with "one-hot" codebooks, such as those generated
    by in-situ sequencing, which guarantee that only one channel will be "on" per round.

    """
    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        image_stack = ctx.obj["image_stack"]
        blobs_stack = ctx.obj["blobs_stack"]
        blobs_axes = ctx.obj["blobs_axes"]

        intensities: IntensityTable = instance.run(
            image_stack,
            blobs_stack,
            blobs_axes,
        )

        # TODO ambrosejcarr find a way to save arbitrary detector results
        intensities.to_netcdf(output)

    @staticmethod
    @click.group("DetectSpots")
    @click.option("-i", "--input", required=True, type=ImageStackParamType)
    @click.option("-o", "--output", required=True)
    @click.option(
        "--blobs-stack",
        default=None,
        required=False,
        type=ImageStackParamType,
        help="ImageStack that contains the blobs."
    )
    @click.option(
        "--blobs-axis",
        type=click.Choice([Axes.ROUND.value, Axes.CH.value, Axes.ZPLANE.value]),
        multiple=True,
        required=False,
        help="The axes that the blobs image will be maj-projected to produce the blobs_image"
    )
    @click.pass_context
    def _cli(ctx, input, output, blobs_stack, blobs_axis):
        """detect spots"""
        print('Detecting Spots ...')
        _blobs_axes = tuple(Axes(_blobs_axis) for _blobs_axis in blobs_axis)

        ctx.obj = dict(
            component=DetectSpots,
            image_stack=input,
            output=output,
            blobs_stack=blobs_stack,
            blobs_axes=_blobs_axes,
        )


class DetectSpotsAlgorithmBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return DetectSpots

    @abstractmethod
    def run(
            self,
            primary_image: ImageStack,
            blobs_image: Optional[ImageStack] = None,
            blobs_axes: Optional[Tuple[Axes, ...]] = None,
            *args,
    ) -> Union[IntensityTable, Tuple[IntensityTable, Any]]:
        """Finds spots in an ImageStack"""
        raise NotImplementedError()

    @abstractmethod
    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """Finds spots in a 3d volume"""
        raise NotImplementedError()

    @staticmethod
    def _get_measurement_function(measurement_type: str) -> Callable[[Sequence], Number]:
        try:
            measurement_function = getattr(np, measurement_type)
        except AttributeError:
            raise ValueError(
                f'measurement_type must be a numpy reduce function such as "max" or "mean". '
                f'{measurement_type} not found.')
        return measurement_function



import json
import os
from typing import Any, Dict, Sequence, Union


class NestedDict(dict):

    def __missing__(self, key):
        self[key] = NestedDict()
        return self[key]

    def update(self, source):
        for k, v in source.items():
            if isinstance(v, dict):
                # Doesn't handle recursion
                self[k] = NestedDict()
                self[k].update(v)
            else:
                self[k] = v


class Config(object):

    __NO_VALUE_PASSED = object()

    def __init__(self, value: Union[str, Dict]=None) -> None:
        """
        Parse user arguments, environment variables, and external files to
        generate a nested configuration object.

        Parameters
        ----------
        value: Union[str, Dict]
            Either a json-object-like structure which will be passed unmodified to the
            Config constructor, or a string which will be used to build such a structure. If
            the string starts with an "@", then it will be interpreted as the filename of a
            json file which should be read. Otherwise, it will be parsed as a json string.
        """
        # Record the original values
        self.__value = value

        if not value:
            value = {}

        data = NestedDict()
        if isinstance(value, str):
            if value.startswith("@"):
                filename = os.path.expanduser(value[1:])
                if os.path.exists(filename):
                    with open(filename, "r") as o:
                        data.update(json.loads(o.read()))
            else:
                data.update(json.loads(value))
        else:
            data.update(value)

        self.data = data

    def lookup(self, keys: Sequence[str],
               value: Any=__NO_VALUE_PASSED,
               remove: bool=False) -> Any:
        """
        Parameters
        ----------
        keys: Sequence[str]
            Sequence of keys which will be looked up one after the other.
            If any key does not exist or None is returned, then the provided
            default value will be returned. If no default value is provided,
            an exception is raised.
        value: Any
            Default value to return in the case that the lookup fails. An
            exception will be raised if none is provided.
        remove: bool
            If True, then delete the key from the object once it has been
            looked up. If parent objects are left with no values, they will
            be removed.
        """
        data: Any = self.data
        stack = []
        for key in keys:
            try:
                stack.append(data)
                data = data.get(key)
            except (AttributeError, KeyError):
                data = None  # Clear value
                if value is Config.__NO_VALUE_PASSED:
                    raise
        if data is not None:
            # If we've reached here without exception,
            # then we've found the value.
            if remove:
                for idx in reversed(range(len(keys))):
                    key = keys[idx]
                    source = stack[idx]
                    source.pop(key)
                    if source:
                        break
            return data
        elif value is not Config.__NO_VALUE_PASSED:
            return value
        else:
            raise KeyError(keys)



from copy import deepcopy
from typing import Optional, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Clip
from starfish.core.util import click
from starfish.core.util.dtype import preserve_float_range
from ._base import FilterAlgorithmBase


class ElementWiseMultiply(FilterAlgorithmBase):
    """
    Perform element-wise multiplication on the image tensor. This is useful for
    performing operations such as image normalization or field flatness correction

    Parameters
    ----------
    mult_mat : xr.DataArray
        the image is element-wise multiplied with this array
    clip_method : Union[str, Clip]
        (Default Clip.CLIP) Controls the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].
        Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
        Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
        value calculated over the entire ImageStack
    """

    def __init__(
        self, mult_array: xr.DataArray, clip_method: Union[str, Clip] = Clip.CLIP
    ) -> None:

        self.mult_array = mult_array
        if clip_method == Clip.SCALE_BY_CHUNK:
            raise ValueError("`scale_by_chunk` is not a valid clip_method for ElementWiseMultiply")
        self.clip_method = clip_method

    _DEFAULT_TESTING_PARAMETERS = {
        "mult_array": xr.DataArray(
            np.array([[[[[1]]], [[[0.5]]]]]),
            dims=('r', 'c', 'z', 'y', 'x')
        )
    }

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        # Align the axes of the multipliers with ImageStack
        mult_array_aligned: np.ndarray = self.mult_array.transpose(*stack.xarray.dims).values
        if not in_place:
            stack = deepcopy(stack)

        stack._data.data.values *= mult_array_aligned
        if self.clip_method == Clip.CLIP:
            stack._data.data.values = preserve_float_range(stack._data.data.values, rescale=False)
        else:
            stack._data.data.values = preserve_float_range(stack._data.data.values, rescale=True)
        return stack

    @staticmethod
    @click.command("ElementWiseMultiply")
    @click.option(
        "--mult-array", required=True, type=np.ndarray, help="matrix to multiply with the image")
    @click.option(
        "--clip-method", default=Clip.CLIP, type=Clip,
        help="method to constrain data to [0,1]. options: 'clip', 'scale_by_image'")
    def _cli(ctx, mult_array, clip_method):
        ctx.obj["component"]._cli_run(ctx, ElementWiseMultiply(mult_array), clip_method)



import mmap
from multiprocessing.heap import Arena
from sys import platform


def anonymous_arena_init(self, size, fd=-1):
    """
    Patching multiprocessing.heap.Arena.__init__ so that it uses anonymous memory mapping
    instead of creating a temp file for shared memory amongst subprocess.

    fix from: https://bugs.python.org/issue30919
    """
    self.size = size
    self.fd = fd  # still kept but is not used !
    self.buffer = mmap.mmap(-1, self.size)


if platform.startswith("darwin"):
    # Only do in Mac env.
    Arena.__init__ = anonymous_arena_init



import importlib
from enum import Enum
from typing import (
    Callable,
    cast,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Union
)

import numpy as np

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes, Clip, Coordinates, Number
from starfish.core.util import click
from starfish.core.util.dtype import preserve_float_range
from ._base import FilterAlgorithmBase


class Reduce(FilterAlgorithmBase):
    """
    Reduces the cardinality of one or more axes to 1 by applying a function across those axes.

    Parameters
    ----------
    dims : Axes
        one or more Axes to reduce over
    func : str
        Name of a function in the module specified by the ``module`` parameter to apply across the
        dimension(s) specified by dims.  The function is resolved by ``getattr(<module>, func)``,
        except in the cases of predefined aliases.  See :py:class:`FunctionSource` for more
        information about aliases.

        Some common examples for the np FunctionSource:

        - amax: maximum intensity projection (applies np.amax)
        - max: maximum intensity projection (this is an alias for amax and applies np.amax)
        - mean: take the mean across the dim(s) (applies np.mean)
        - sum: sum across the dim(s) (applies np.sum)
    module : FunctionSource
        Python module that serves as the source of the function.  It must be listed as one of the
        members of :py:class:`FunctionSource`.

        Currently, the supported FunctionSources are:
        - ``np``: the top-level package of numpy
        - ``scipy``: the top-level package of scipy
    clip_method : Clip
        (Default Clip.CLIP) Controls the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].
        Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
        Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
        value calculated over the entire ImageStack

    Examples
    --------
    Reducing via max projection.
        >>> from starfish.core.imagestack.test.factories.synthetic_stack import synthetic_stack
        >>> from starfish.image import Filter
        >>> from starfish.types import Axes
        >>> stack = synthetic_stack()
        >>> reducer = Filter.Reduce({Axes.ROUND}, func="max")
        >>> max_proj = reducer.run(stack)

    Reducing via linalg.norm
        >>> from starfish.core.imagestack.test.factories.synthetic_stack import synthetic_stack
        >>> from starfish.image import Filter
        >>> from starfish.types import Axes
        >>> stack = synthetic_stack()
        >>> reducer = Filter.Reduce(
                {Axes.ROUND},
                func="linalg.norm",
                module=Filter.Reduce.FunctionSource.scipy,
                ord=2,
            )
        >>> norm = reducer.run(stack)

    See Also
    --------
    starfish.core.types.Axes

    """

    class FunctionSource(Enum):
        """Each FunctionSource declares a package from which reduction methods can be obtained.
        Generally, the packages should be those that are included as starfish's dependencies for
        reproducibility.

        Many packages are broken into subpackages which are not necessarily implicitly imported when
        importing the top-level package.  For example, ``scipy.linalg`` is not implicitly imported
        when one imports ``scipy``.  To avoid the complexity of enumerating each scipy subpackage in
        FunctionSource, we assemble the fully-qualified method name, and then try all the
        permutations of how one could import that method.

        In the example of ``scipy.linalg.norm``, we try the following:

        1. import ``scipy``, attempt to resolve ``linalg.norm``.
        2. import ``scipy.linalg``, attempt to resolve ``norm``.
        """

        def __init__(self, top_level_package: str, aliases: Optional[Mapping[str, str]] = None):
            self.top_level_package = top_level_package
            self.aliases = aliases or {}

        def _resolve_method(self, method: str) -> Callable:
            """Resolve a method.  The method itself might be enclosed in a package, such as
            subpackage.actual_method.  In that case, we will need to attempt to resolve it in the
            following sequence:

            1. import top_level_package, then try to resolve subpackage.actual_method recursively
               through ``getattr`` calls.
            2. import top_level_package.subpackage, then try to resolve actual_method through
               ``gettatr`` calls.

            This is done instead of just creating a bunch of FunctionSource for libraries that have
            a lot of packages that are not implicitly imported by importing the top-level package.
            """
            # first resolve the aliases.
            actual_method = self.aliases.get(method, method)

            method_splitted = actual_method.split(".")
            splitted = [self.top_level_package]
            splitted.extend(method_splitted)

            for divider in range(1, len(splitted)):
                import_section = splitted[:divider]
                getattr_section = splitted[divider:]

                imported = importlib.import_module(".".join(import_section))

                try:
                    for getattr_name in getattr_section:
                        imported = getattr(imported, getattr_name)
                    return cast(Callable, imported)
                except AttributeError:
                    pass

            raise AttributeError(
                f"Unable to resolve the method {actual_method} from package "
                f"{self.top_level_package}")

        np = ("numpy", {'max': 'amax'})
        """Function source for the numpy libraries"""
        scipy = ("scipy",)

    def __init__(
        self,
            dims: Iterable[Union[Axes, str]],
            func: str = "max",
            module: FunctionSource = FunctionSource.np,
            clip_method: Clip = Clip.CLIP,
            **kwargs
    ) -> None:
        self.dims = dims
        self.func = module._resolve_method(func)
        self.clip_method = clip_method
        self.kwargs = kwargs

    _DEFAULT_TESTING_PARAMETERS = {"dims": ['r'], "func": 'max'}

    def run(
            self,
            stack: ImageStack,
            *args,
    ) -> ImageStack:
        """Performs the dimension reduction with the specifed function

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack. Otherwise return the
            original stack.

        """

        # Apply the reducing function
        reduced = stack._data.reduce(
            self.func, dim=[Axes(dim).value for dim in self.dims], **self.kwargs)

        # Add the reduced dims back and align with the original stack
        reduced = reduced.expand_dims(tuple(Axes(dim).value for dim in self.dims))
        reduced = reduced.transpose(*stack.xarray.dims)

        if self.clip_method == Clip.CLIP:
            reduced = preserve_float_range(reduced, rescale=False)
        else:
            reduced = preserve_float_range(reduced, rescale=True)

        # Update the physical coordinates
        physical_coords: MutableMapping[Coordinates, Sequence[Number]] = {}
        for axis, coord in (
                (Axes.X, Coordinates.X),
                (Axes.Y, Coordinates.Y),
                (Axes.ZPLANE, Coordinates.Z)):
            if axis in self.dims:
                # this axis was projected out of existence.
                assert coord.value not in reduced.coords
                physical_coords[coord] = [np.average(stack._data.coords[coord.value])]
            else:
                physical_coords[coord] = reduced.coords[coord.value]
        reduced_stack = ImageStack.from_numpy(reduced.values, coordinates=physical_coords)

        return reduced_stack

    @staticmethod
    @click.command("Reduce")
    @click.option(
        "--dims",
        type=click.Choice(
            [Axes.ROUND.value, Axes.CH.value, Axes.ZPLANE.value, Axes.X.value, Axes.Y.value]
        ),
        multiple=True,
        help="The dimensions the Imagestack should max project over."
             "For multiple dimensions add multiple --dims. Ex."
             "--dims r --dims c")
    @click.option(
        "--func",
        type=str,
        help="The function to apply across dims."
    )
    @click.option(
        "--module",
        type=click.Choice([member.name for member in list(FunctionSource)]),
        multiple=False,
        help="Module to source the function from.",
        default=FunctionSource.np.name,
    )
    @click.option(
        "--clip-method", default=Clip.CLIP, type=Clip,
        help="method to constrain data to [0,1]. options: 'clip', 'scale_by_image'")
    @click.pass_context
    def _cli(ctx, dims, func, module, clip_method):
        ctx.obj["component"]._cli_run(
            ctx, Reduce(dims, func, Reduce.FunctionSource[module], clip_method))



import numpy as np
from scipy.ndimage.filters import gaussian_filter

from starfish import ImageStack
from starfish.core.spots._detect_spots.local_search_blob_detector import LocalSearchBlobDetector
from starfish.core.types import Axes


def traversing_code() -> ImageStack:
    """this code walks in a sequential direction, and should only be detectable from some anchors"""
    img = np.zeros((3, 2, 20, 50, 50), dtype=np.float32)

    # code 1
    img[0, 0, 5, 35, 35] = 10
    img[1, 1, 5, 32, 32] = 10
    img[2, 0, 5, 29, 29] = 10

    # blur points
    gaussian_filter(img, (0, 0, 0.5, 1.5, 1.5), output=img)

    return ImageStack.from_numpy(img)


def multiple_possible_neighbors() -> ImageStack:
    """this image is intended to be tested with anchor_round in {0, 1}, last round has more spots"""
    img = np.zeros((3, 2, 20, 50, 50), dtype=np.float32)

    # round 1
    img[0, 0, 5, 20, 40] = 10
    img[0, 0, 5, 40, 20] = 10

    # round 2
    img[1, 1, 5, 20, 40] = 10
    img[1, 1, 5, 40, 20] = 10

    # round 3
    img[2, 0, 5, 20, 40] = 10
    img[2, 0, 5, 35, 35] = 10
    img[2, 0, 5, 40, 20] = 10

    # blur points
    gaussian_filter(img, (0, 0, 0.5, 1.5, 1.5), output=img)

    return ImageStack.from_numpy(img)


def jitter_code() -> ImageStack:
    """this code has some minor jitter <= 3px at the most distant point"""
    img = np.zeros((3, 2, 20, 50, 50), dtype=np.float32)

    # code 1
    img[0, 0, 5, 35, 35] = 10
    img[1, 1, 5, 34, 35] = 10
    img[2, 0, 6, 35, 33] = 10

    # blur points
    gaussian_filter(img, (0, 0, 0.5, 1.5, 1.5), output=img)

    return ImageStack.from_numpy(img)


def two_perfect_codes() -> ImageStack:
    """this code has no jitter"""
    img = np.zeros((3, 2, 20, 50, 50), dtype=np.float32)

    # code 1
    img[0, 0, 5, 20, 35] = 10
    img[1, 1, 5, 20, 35] = 10
    img[2, 0, 5, 20, 35] = 10

    # code 1
    img[0, 0, 5, 40, 45] = 10
    img[1, 1, 5, 40, 45] = 10
    img[2, 0, 5, 40, 45] = 10

    # blur points
    gaussian_filter(img, (0, 0, 0.5, 1.5, 1.5), output=img)

    return ImageStack.from_numpy(img)


def local_search_blob_detector(search_radius: int, anchor_channel=0) -> LocalSearchBlobDetector:
    return LocalSearchBlobDetector(
        min_sigma=(0.4, 1.2, 1.2),
        max_sigma=(0.6, 1.7, 1.7),
        num_sigma=3,
        threshold=0.1,
        overlap=0.5,
        search_radius=search_radius,
        anchor_round=anchor_channel,
    )


def test_local_search_blob_detector_two_codes():
    stack = two_perfect_codes()
    lsbd = local_search_blob_detector(search_radius=1)
    intensity_table = lsbd.run(stack, n_processes=1)

    assert intensity_table.shape == (2, 2, 3)
    assert np.all(intensity_table[0][Axes.X.value] == 45)
    assert np.all(intensity_table[0][Axes.Y.value] == 40)
    assert np.all(intensity_table[0][Axes.ZPLANE.value] == 5)


def test_local_search_blob_detector_jitter_code():
    stack = jitter_code()
    lsbd = local_search_blob_detector(search_radius=3)
    intensity_table = lsbd.run(stack, n_processes=1)

    assert intensity_table.shape == (1, 2, 3)
    f, c, r = np.where(~intensity_table.isnull())
    assert np.all(f == np.array([0, 0, 0]))
    assert np.all(c == np.array([0, 0, 1]))
    assert np.all(r == np.array([0, 2, 1]))

    # test again with smaller search radius
    lsbd = local_search_blob_detector(search_radius=1)
    intensity_table = lsbd.run(stack, n_processes=1)
    assert intensity_table.shape == (1, 2, 3)
    f, c, r = np.where(~intensity_table.isnull())
    assert np.all(f == np.array([0]))
    assert np.all(c == np.array([0]))
    assert np.all(r == np.array([0]))


def test_local_search_blob_detector_traversing_code():
    stack = traversing_code()
    lsbd = local_search_blob_detector(search_radius=5, anchor_channel=0)
    intensity_table = lsbd.run(stack, n_processes=1)

    assert intensity_table.shape == (1, 2, 3)
    f, c, r = np.where(~intensity_table.isnull())
    assert np.all(f == np.array([0, 0]))
    assert np.all(c == np.array([0, 1]))
    assert np.all(r == np.array([0, 1]))

    lsbd = local_search_blob_detector(search_radius=5, anchor_channel=1)
    intensity_table = lsbd.run(stack, n_processes=1)
    f, c, r = np.where(~intensity_table.isnull())
    assert np.all(f == np.array([0, 0, 0]))
    assert np.all(c == np.array([0, 0, 1]))
    assert np.all(r == np.array([0, 2, 1]))


def test_local_search_blob_detector_multiple_neighbors():
    stack = multiple_possible_neighbors()
    lsbd = local_search_blob_detector(search_radius=4, anchor_channel=0)
    intensity_table = lsbd.run(stack, n_processes=1)

    assert intensity_table.shape == (2, 2, 3)
    f, c, r = np.where(~intensity_table.isnull())
    assert np.all(intensity_table[Axes.ZPLANE.value] == (5, 5))
    assert np.all(intensity_table[Axes.Y.value] == (40, 20))
    assert np.all(intensity_table[Axes.X.value] == (20, 40))



from typing import Optional, Sequence

from starfish.core.experiment.builder.test.factories.unique_tiles import UniqueTiles
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.imagestack.parser.crop import CropParameters
from .all_purpose import imagestack_factory

X_COORDS = 0.01, 0.1
Y_COORDS = 0.001, 0.01
Z_COORDS = 0.0001, 0.001


def unique_tiles_imagestack(
        round_labels: Sequence[int],
        ch_labels: Sequence[int],
        zplane_labels: Sequence[int],
        tile_height: int,
        tile_width: int,
        crop_parameters: Optional[CropParameters] = None) -> ImageStack:
    """Build an imagestack with unique values per tile.
    """
    return imagestack_factory(
        UniqueTiles,
        round_labels,
        ch_labels,
        zplane_labels,
        tile_height,
        tile_width,
        X_COORDS,
        Y_COORDS,
        Z_COORDS,
        crop_parameters,
    )



from starfish.core.experiment.builder.orderediterator import ordered_iterator
from starfish.core.types import Axes


def test_round_then_ch():
    results = list(ordered_iterator(((Axes.ROUND, range(3)), (Axes.CH, range(2)))))
    for ix in (0, 1):
        assert results[ix][Axes.ROUND] == 0
    for ix in (2, 3):
        assert results[ix][Axes.ROUND] == 1
    for ix in (4, 5):
        assert results[ix][Axes.ROUND] == 2

    for ix in (0, 2, 4):
        assert results[ix][Axes.CH] == 0
    for ix in (1, 3, 5):
        assert results[ix][Axes.CH] == 1


def test_ch_then_round():
    results = list(ordered_iterator(((Axes.CH, range(2)), (Axes.ROUND, range(3)))))
    for ix in (0, 3):
        assert results[ix][Axes.ROUND] == 0
    for ix in (1, 4):
        assert results[ix][Axes.ROUND] == 1
    for ix in (2, 5):
        assert results[ix][Axes.ROUND] == 2

    for ix in (0, 1, 2):
        assert results[ix][Axes.CH] == 0
    for ix in (3, 4, 5):
        assert results[ix][Axes.CH] == 1



import numpy as np
import pytest
import xarray as xr

from starfish.core.intensity_table.test.factories import synthetic_intensity_table
from starfish.core.test.factories import synthetic_spot_pass_through_stack, SyntheticData
from starfish.core.types import Axes


def test_synthetic_spot_creation_raises_error_with_coords_too_small():
    num_z = 0
    height = 40
    width = 50
    intensity_table = synthetic_intensity_table()
    with pytest.raises(ValueError):
        SyntheticData.synthetic_spots(intensity_table, num_z, height, width)


def test_synthetic_spot_creation_produces_an_imagestack_with_correct_spot_location():

    codebook, true_intensities, image = synthetic_spot_pass_through_stack()

    g, c, r = np.where(true_intensities.values)

    x = np.empty_like(g)
    y = np.empty_like(g)
    z = np.empty_like(g)
    breaks = np.concatenate([
        np.array([0]),
        np.where(np.diff(g))[0] + 1,
        np.array([g.shape[0]])
    ])
    for i in np.arange(len(breaks) - 1):
        x[breaks[i]: breaks[i + 1]] = true_intensities.coords[Axes.X.value][i]
        y[breaks[i]: breaks[i + 1]] = true_intensities.coords[Axes.Y.value][i]
        z[breaks[i]: breaks[i + 1]] = true_intensities.coords[Axes.ZPLANE.value][i]

    # only 8 values should be set, since there are only 8 locations across the tensor
    assert np.sum(image.xarray != 0) == 8

    intensities = image.xarray.sel(
        x=xr.DataArray(x, dims=['intensity']),
        y=xr.DataArray(y, dims=['intensity']),
        z=xr.DataArray(z, dims=['intensity']),
        r=xr.DataArray(r, dims=['intensity']),
        c=xr.DataArray(c, dims=['intensity']))
    assert np.allclose(
        intensities,
        true_intensities.values[np.where(true_intensities)])



from collections import OrderedDict

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.imagestack.physical_coordinates import _get_physical_coordinates_of_z_plane
from starfish.core.types import Axes, Coordinates, PhysicalCoordinateTypes
from .synthetic_stack import synthetic_stack


def imagestack_with_coords_factory(stack_shape: OrderedDict, coords: OrderedDict) -> ImageStack:
    """
    Create an ImageStack of given shape and assigns the given x,y,z
    min/max physical coordinates to each tile.

    Parameters
    ----------
    stack_shape: OrderedDict
        Dict[Axes, int] defining the size of each dimension for an ImageStack

    coords: OrderedDict
        Dict[PhysicalCoordinateTypes, float] defining the min/max values of physical
        coordinates to assign to the Imagestack
    """

    stack = synthetic_stack(
        num_round=stack_shape[Axes.ROUND],
        num_ch=stack_shape[Axes.CH],
        num_z=stack_shape[Axes.ZPLANE],
        tile_height=stack_shape[Axes.Y],
        tile_width=stack_shape[Axes.X],
    )

    stack.xarray[Coordinates.X.value] = xr.DataArray(
        np.linspace(coords[PhysicalCoordinateTypes.X_MIN], coords[PhysicalCoordinateTypes.X_MAX],
                    stack.xarray.sizes[Axes.X.value]), dims=Axes.X.value)

    stack.xarray[Coordinates.Y.value] = xr.DataArray(
        np.linspace(coords[PhysicalCoordinateTypes.Y_MIN], coords[PhysicalCoordinateTypes.Y_MAX],
                    stack.xarray.sizes[Axes.Y.value]), dims=Axes.Y.value)

    z_coord = _get_physical_coordinates_of_z_plane(
        (coords[PhysicalCoordinateTypes.Z_MIN], coords[PhysicalCoordinateTypes.Z_MAX]))

    stack.xarray[Coordinates.Z.value] = xr.DataArray(np.zeros(
        stack.xarray.sizes[Axes.ZPLANE.value]),
        dims=Axes.ZPLANE.value)

    for z in stack.axis_labels(Axes.ZPLANE):
        stack.xarray[Coordinates.Z.value].loc[z] = z_coord

    return stack



from abc import abstractmethod
from typing import Type


from starfish.core.image._registration.transforms_list import TransformsList
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.pipeline.pipelinecomponent import PipelineComponent
from starfish.core.util import click
from starfish.core.util.click.indirectparams import ImageStackParamType


class LearnTransform(PipelineComponent):
    """
    LearnTransform exposes methods to learn transformations that align ImageStacks.
    """
    @classmethod
    def _cli_run(cls, ctx, instance, *args, **kwargs):
        output = ctx.obj["output"]
        stack = ctx.obj["stack"]
        transformation_list = instance.run(stack)
        transformation_list.to_json(output)

    @staticmethod
    @click.group("LearnTransform")
    @click.option("-i", "--input", type=ImageStackParamType)
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, input, output):
        """Learn a set of transforms for an ImageStack."""
        print("Learning Transforms for images...")
        ctx.obj = dict(
            component=LearnTransform,
            output=output,
            stack=input,
        )


class LearnTransformBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return LearnTransform

    @abstractmethod
    def run(self, stack, *args) -> TransformsList:
        """Learns Transforms for a given stack."""
        raise NotImplementedError()



"""
This module parses and retains the extras metadata attached to TileSet extras.
"""
from typing import Collection, Mapping, MutableMapping, Sequence, Tuple

import numpy as np
from slicedimage import Tile, TileSet

from starfish.core.imagestack.dataorder import AXES_DATA
from starfish.core.imagestack.parser import TileCollectionData, TileData, TileKey
from starfish.core.imagestack.physical_coordinates import _get_physical_coordinates_of_z_plane
from starfish.core.types import Axes, Coordinates, Number


class SlicedImageTile(TileData):
    """
    This wraps a :py:class:`slicedimage.Tile`.  The difference between this and
    :py:class:`slicedimage.Tile` is that this class does cache the image data upon load.  It is
    therefore incumbent on the consumers of these objects to discard them as soon as it is
    reasonable to do so to free up memory.
    """
    def __init__(
            self,
            wrapped_tile: Tile,
            r: int,
            ch: int,
            zplane: int,
    ) -> None:
        self._wrapped_tile = wrapped_tile
        self._r = r
        self._ch = ch
        self._zplane = zplane
        self._numpy_array: np.ndarray = None

    def _load(self):
        if self._numpy_array is not None:
            return
        self._numpy_array = self._wrapped_tile.numpy_array

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        self._load()
        raw_tile_shape = self._numpy_array.shape
        assert len(raw_tile_shape) == 2
        tile_shape = {Axes.Y: raw_tile_shape[0], Axes.X: raw_tile_shape[1]}
        return tile_shape

    @property
    def numpy_array(self) -> np.ndarray:
        self._load()
        return self._numpy_array

    @property
    def coordinates(self) -> Mapping[Coordinates, Sequence[Number]]:
        xrange = self._wrapped_tile.coordinates[Coordinates.X]
        yrange = self._wrapped_tile.coordinates[Coordinates.Y]
        return_coords = {
            Coordinates.X: np.linspace(xrange[0], xrange[1], self.tile_shape[Axes.X]),
            Coordinates.Y: np.linspace(yrange[0], yrange[1], self.tile_shape[Axes.Y]),
        }

        if Coordinates.Z in self._wrapped_tile.coordinates:
            zrange = self._wrapped_tile.coordinates[Coordinates.Z]
            zplane_coord = _get_physical_coordinates_of_z_plane(zrange)
            return_coords[Coordinates.Z] = [zplane_coord]

        return return_coords

    @property
    def selector(self) -> Mapping[Axes, int]:
        return {
            Axes(axis_str): index
            for axis_str, index in self._wrapped_tile.indices.items()
        }


class TileSetData(TileCollectionData):
    """
    This class presents a simpler API for accessing a TileSet and its constituent tiles.
    """
    def __init__(self, tileset: TileSet) -> None:
        self._tile_shape = tileset.default_tile_shape

        self.tiles: MutableMapping[TileKey, Tile] = dict()
        for tile in tileset.tiles():
            key = TileKey(
                round=tile.indices[Axes.ROUND],
                ch=tile.indices[Axes.CH],
                zplane=tile.indices.get(Axes.ZPLANE, 0))
            self.tiles[key] = tile

            # if we don't have the tile shape, then we peek at the tile and get its shape.
            if self._tile_shape is None:
                self._tile_shape = tile.tile_shape

        self._extras = tileset.extras

    def __getitem__(self, tilekey: TileKey) -> dict:
        """Returns the extras metadata for a given tile, addressed by its TileKey"""
        return self.tiles[tilekey].extras

    def keys(self) -> Collection[TileKey]:
        """Returns a Collection of the TileKey's for all the tiles."""
        return self.tiles.keys()

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        return self._tile_shape

    @property
    def extras(self) -> dict:
        """Returns the extras metadata for the TileSet."""
        return self._extras

    def get_tile_by_key(self, tilekey: TileKey) -> TileData:
        return SlicedImageTile(
            self.tiles[tilekey],
            tilekey.round, tilekey.ch, tilekey.z,
        )

    def get_tile(self, r: int, ch: int, z: int) -> TileData:
        return SlicedImageTile(
            self.tiles[TileKey(round=r, ch=ch, zplane=z)],
            r, ch, z,
        )


def parse_tileset(
        tileset: TileSet
) -> Tuple[Mapping[Axes, int], TileCollectionData]:
    """
    Parse a :py:class:`slicedimage.TileSet` for formatting into an
    :py:class:`starfish.imagestack.ImageStack`.

    Parameters:
    -----------
    tileset : TileSet
        The tileset to parse.

    Returns:
    --------
    Tuple[Tuple[int, int], TileSetData] :
        A tuple consisting of the following:
            1. The (y, x) size of each tile.
            2. A :py:class:`starfish.imagestack.tileset.TileSetData` that can be queried to obtain
               the image data and extras metadata of each tile, as well as the extras metadata of
               the entire :py:class:`slicedimage.TileSet`.
    """
    tile_data = TileSetData(tileset)

    tile_shape = tileset.default_tile_shape

    # if we don't have the tile shape, then we peek at the first tile and get its shape.
    if tile_shape is None:
        tile_key = next(iter(tile_data.keys()))
        tile = tile_data.get_tile_by_key(tile_key)
        tile_shape = tile.tile_shape

    return (
        tile_shape,
        tile_data,
    )


def _get_dimension_size(tileset: TileSet, dimension: Axes):
    axis_data = AXES_DATA[dimension]
    if dimension in tileset.dimensions or axis_data.required:
        return tileset.get_dimension_shape(dimension)
    return 1



from enum import Enum


class AugmentedEnum(Enum):
    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, other):
        if isinstance(other, type(self)) or isinstance(other, str):
            return self.value == other
        return False

    def __str__(self) -> str:
        return self.value


class Coordinates(AugmentedEnum):
    Z = 'zc'
    Y = 'yc'
    X = 'xc'


PHYSICAL_COORDINATE_DIMENSION = "physical_coordinate"

"""
This is the xarray dimension name for the physical coordinates of the tiles.
"""

STARFISH_EXTRAS_KEY = 'starfish'
"""
Attribute on Imagestack and IntensityTable for storing starfish related info
"""
LOG = "log"
"""
This is name of the provenance log attribute stored on the IntensityTable
"""
CORE_DEPENDENCIES = {'numpy', 'scikit-image', 'pandas', 'scikit-learn', 'scipy', 'xarray', 'sympy'}
"""
The set of dependencies whose versions are are logged for each starfish session
"""


class PhysicalCoordinateTypes(AugmentedEnum):
    """
    These are more accurately the xarray coordinates for the physical coordinates of a tile.
    """
    Z_MAX = 'zmax'
    Z_MIN = 'zmin'
    Y_MAX = 'ymax'
    Y_MIN = 'ymin'
    X_MAX = 'xmax'
    X_MIN = 'xmin'


class Axes(AugmentedEnum):
    ROUND = 'r'
    CH = 'c'
    ZPLANE = 'z'
    Y = 'y'
    X = 'x'


class Features:
    """
    contains constants relating to the codebook and feature (spot/pixel) representations of the
    image data
    """

    AXIS = 'features'
    TARGET = 'target'
    CODEWORD = 'codeword'
    CODE_VALUE = 'v'
    SPOT_RADIUS = 'radius'
    DISTANCE = 'distance'
    PASSES_THRESHOLDS = 'passes_thresholds'
    CELL_ID = 'cell_id'
    SPOT_ID = 'spot_id'
    INTENSITY = 'intensity'
    AREA = 'area'
    CELLS = 'cells'
    GENES = 'genes'


class OverlapStrategy(AugmentedEnum):
    """
    contains options to use when processes physically overlapping IntensityTables
    or ImageStacks
    """
    TAKE_MAX = 'take_max'


class Clip(AugmentedEnum):
    """
    contains clipping options that determine how out-of-bounds values produced by filters are
    treated to keep the image contained within [0, 1]
    """
    CLIP = 'clip'
    SCALE_BY_IMAGE = 'scale_by_image'
    SCALE_BY_CHUNK = 'scale_by_chunk'


class TransformType(AugmentedEnum):
    """
    currently supported transform types
    """
    SIMILARITY = 'similarity'



import abc
from typing import Generic, Iterable, TypeVar


FormulaResultType = TypeVar("FormulaResultType")


class ConversionFormula(Generic[FormulaResultType]):
    """A conversion formula represents a plausible contract to convert a string value to another
    data type.  Each conversion formula implements two methods -- a lightweight method to determine
    if this formula can be applied given a string value, and a load method that actually does the
    conversion.
    """
    @abc.abstractmethod
    def applicable(self, input_parameter: str) -> bool:
        """Returns true iff this formula might work for the given input."""
        raise NotImplementedError()

    @abc.abstractmethod
    def load(self, input_parameter: str) -> FormulaResultType:
        """Attempt to run this conversion formula against this input."""
        raise NotImplementedError()


class NoApplicableConversionFormulaError(Exception):
    """Raised when no conversion formula declared itself applicable to this input string."""
    pass


class NoSuccessfulConversionFormulaError(Exception):
    """Raised when all the conversion formulas that declared itself applicable failed to execute
    successfully."""
    pass


ConvertResultType = TypeVar("ConvertResultType")


def convert(
        value: str,
        conversion_formulas: Iterable[ConversionFormula[ConvertResultType]]) -> ConvertResultType:
    """
    Given a string value and a series of conversion formulas, attempt to convert the value using the
    formulas.

    If none of the formulas declare themselves as applicable, then raise
    :py:class:`NoApplicableConversionformulaError`.  If none of the formulas that declare themselves
    as eligible run successfully, then raise :py:class:`NoSuccessfulConversionformulaError`.

    Parameters
    ----------
    value : str
        The string value we are attempting to convert.
    conversion_formulas : Iterable[ConversionFormula[ConvertResultType]]
        A series of conversion formulas.

    Returns
    -------
    The converted value.
    """
    none_applied = True

    for conversion_formula in conversion_formulas:
        if conversion_formula.applicable(value):
            none_applied = False
            try:
                return conversion_formula.load(value)
            except Exception:
                pass

    if none_applied:
        raise NoApplicableConversionFormulaError(
            f"Could not find applicable gonversion formula for {value}")
    raise NoSuccessfulConversionFormulaError(
        f"All applicable conversion formulas failed to run successfully for {value}.")



"""
Notes
-----
This test and docs/source/usage/iss/iss_cli.sh test the same code paths and should be updated
together
"""
import os
import unittest
from pathlib import Path
from typing import Iterable

import numpy as np
import pandas as pd
import pytest

from starfish import IntensityTable
from starfish.types import Features
from ._base_recipe_test import RecipeTest


URL = "https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/experiment.json"


@pytest.mark.slow
class TestWithIssData(RecipeTest, unittest.TestCase):
    @property
    def recipe(self) -> Path:
        test_file_path = Path(__file__)
        recipe = test_file_path.parent / "iss_recipe.txt"
        return recipe

    @property
    def input_url_or_paths(self) -> Iterable[str]:
        return [
            f"@{URL}[fov_001][primary]",  # primary image
            f"@{URL}[fov_001][dots]",     # dots image
            f"@{URL}[fov_001][nuclei]",   # nuclei image
            f"@{URL}",                    # codebook
        ]

    @property
    def output_paths(self) -> Iterable[Path]:
        return [
            Path("decoded_spots.nc")
        ]

    def verify_results(self, tempdir: Path):
        intensities = IntensityTable.open_netcdf(os.fspath(tempdir / "decoded_spots.nc"))
        genes, counts = np.unique(
            intensities.coords[Features.TARGET], return_counts=True)
        gene_counts = pd.Series(counts, genes)
        assert gene_counts['ACTB'] == 9
        assert gene_counts['GAPDH'] == 9



import functools
import inspect
from abc import ABCMeta, abstractmethod
from typing import Type

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import LOG
from starfish.core.types._constants import STARFISH_EXTRAS_KEY
from starfish.core.util.logging import LogEncoder
from .pipelinecomponent import PipelineComponent


class AlgorithmBaseType(ABCMeta):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if not inspect.isabstract(cls):
            AlgorithmBaseType.register_with_pipeline_component(cls)
            cls.run = AlgorithmBaseType.run_with_logging(cls.run)

    @staticmethod
    def register_with_pipeline_component(algorithm_cls):
        pipeline_component_cls = algorithm_cls.get_pipeline_component_class()
        if pipeline_component_cls._algorithm_to_class_map_int is None:
            pipeline_component_cls._algorithm_to_class_map_int = {}
        pipeline_component_cls._algorithm_to_class_map_int[algorithm_cls.__name__] = algorithm_cls
        setattr(pipeline_component_cls, algorithm_cls._get_algorithm_name(), algorithm_cls)

        pipeline_component_cls._cli.add_command(algorithm_cls._cli)

    @staticmethod
    def run_with_logging(func):
        """
        This method extends each pipeline component.run() method to also log itself and
        runtime parameters to the IntensityTable and ImageStack objects. There are two
        scenarios for this method:
            1.) Filtering/ApplyTransform:
                    Imagestack -> Imagestack
            2.) Spot Detection:
                    ImageStack -> IntensityTable
                    ImageStack -> [IntensityTable, ConnectedComponentDecodingResult]
            TODO segmentation and decoding
        """
        @functools.wraps(func)
        def helper(*args, **kwargs):
            result = func(*args, **kwargs)
            # Scenario 1, Filtering, ApplyTransform
            if isinstance(result, ImageStack):
                result.update_log(args[0])
            # Scenario 2, Spot detection
            elif isinstance(result, tuple) or isinstance(result, IntensityTable):
                if isinstance(args[1], ImageStack):
                    stack = args[1]
                    # update log with spot detection instance args[0]
                    stack.update_log(args[0])
                    # get resulting intensity table and set log
                    it = result
                    if isinstance(result, tuple):
                        it = result[0]
                    it.attrs[STARFISH_EXTRAS_KEY] = LogEncoder().encode({LOG: stack.log})
            return result
        return helper


class AlgorithmBase(metaclass=AlgorithmBaseType):

    """
    This is the base class of any algorithm that starfish exposes.

    Subclasses of this base class are paired with subclasses of PipelineComponent. The subclasses of
    PipelineComponent retrieve subclasses of the paired AlgorithmBase. Together, the two classes
    enable starfish to expose a paired API and CLI.

    Examples
    --------

    PipelineComponent: `starfish.image._segment.Segmentation(PipelineComponent)`

    AlgorithmBase: `starfish.image._segment._base.SegmentationAlgorithmBase(AlgorithmBase)`

    Implementing Algorithms:
    - `starfish.image._segment.watershed.Watershed(SegmentationAlgorithmBase)`

    This pattern exposes the API as follows:

    `starfish.image.Segmentation.<implementing algorithm (Watershed)>`

    and the CLI as:

    `$> starfish segmentation watershed`

    To create an entirely new group of related algorithms, like `Segmentation`, a new subclass of
    both `AlgorithmBase` and `PipelineComponent` must be created.

    To add to an existing group of algorithms like "Segmentation", an algorithm implementation must
    subclass the corresponding subclass of `AlgorithmBase`. In this case,
    `SegmentationAlgorithmBase`.

    See Also
    --------
    starfish.pipeline.pipelinecomponent.py
    """
    @classmethod
    def _get_algorithm_name(cls):
        """
        Returns the name of the algorithm.  This should be a valid python identifier, i.e.,
        https://docs.python.org/3/reference/lexical_analysis.html#identifiers
        """
        return cls.__name__

    @classmethod
    @abstractmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        """
        Returns the class of PipelineComponent this algorithm implements.
        """
        raise NotImplementedError()



import numpy as np

from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.segmentation_mask import SegmentationMaskCollection
from starfish.core.types import Features
from starfish.core.util import click
from ._base import AssignTargetsAlgorithm


class Label(AssignTargetsAlgorithm):
    """
    Extract cell ids for features in IntensityTable from a set of segmentation masks.
    """

    def __init__(self, **kwargs) -> None:
        pass

    @classmethod
    def _add_arguments(cls, parser) -> None:
        pass

    @staticmethod
    def _assign(
        masks: SegmentationMaskCollection,
        intensities: IntensityTable,
        in_place: bool,
    ) -> IntensityTable:

        intensities[Features.CELL_ID] = (
            Features.AXIS,
            np.full(intensities.sizes[Features.AXIS], fill_value='nan', dtype='<U8')
        )

        for mask in masks:
            y_min, y_max = float(mask.y.min()), float(mask.y.max())
            x_min, x_max = float(mask.x.min()), float(mask.x.max())

            in_bbox = intensities.where(
                (intensities.y >= y_min)
                & (intensities.y <= y_max)
                & (intensities.x >= x_min)
                & (intensities.x <= x_max),
                drop=True
            )

            in_mask = mask.sel_points(y=in_bbox.y, x=in_bbox.x)
            spot_ids = in_bbox[Features.SPOT_ID][in_mask.values]
            intensities[Features.CELL_ID].loc[spot_ids] = mask.name

        return intensities

    def run(
            self,
            masks: SegmentationMaskCollection,
            intensity_table: IntensityTable,
            verbose: bool = False,
            in_place: bool = False,
    ) -> IntensityTable:
        """Extract cell ids for features in IntensityTable from a segmentation label image

        Parameters
        ----------
        masks : SegmentationMaskCollection
            binary masks segmenting each cell
        intensity_table : IntensityTable
            spot information
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)

        Returns
        -------
        IntensityTable :
            IntensityTable with added features variable containing cell ids. Points outside of
            cells will be assigned zero.

        """
        return self._assign(masks, intensity_table, in_place=in_place)

    @staticmethod
    @click.command("Label")
    @click.pass_context
    def _cli(ctx):
        ctx.obj["component"]._cli_run(ctx, Label())



from typing import Tuple, Union

from ._constants import (
    Axes,
    Clip,
    Coordinates,
    CORE_DEPENDENCIES,
    Features,
    LOG,
    OverlapStrategy,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
    STARFISH_EXTRAS_KEY,
    TransformType
)
from ._decoded_spots import DecodedSpots
from ._spot_attributes import SpotAttributes

Number = Union[int, float]
CoordinateValue = Union[Number, Tuple[Number, Number]]



from starfish import ImageStack
from starfish.core.image._filter._base import FilterAlgorithmBase
from starfish.core.util import click


class SimpleFilterAlgorithm(FilterAlgorithmBase):
    def __init__(self, multiplicand: float):
        self.multiplicand = multiplicand

    def run(self, stack: ImageStack, *args) -> ImageStack:
        numpy_array = stack.xarray
        numpy_array = numpy_array * self.multiplicand
        return ImageStack.from_numpy(numpy_array)

    @staticmethod
    @click.command("SimpleFilterAlgorithm")
    @click.option(
        "--multiplicand", default=1.0, type=float)
    @click.pass_context
    def _cli(ctx, multiplicand):
        ctx.obj["component"]._cli_run(ctx, SimpleFilterAlgorithm(multiplicand=multiplicand))


class AdditiveFilterAlgorithm(FilterAlgorithmBase):
    def __init__(self, additive: ImageStack):
        self.additive = additive

    def run(self, stack: ImageStack, *args) -> ImageStack:
        numpy_array = stack.xarray
        numpy_array = numpy_array + stack.xarray
        return ImageStack.from_numpy(numpy_array)

    @staticmethod
    @click.command("AdditiveFilterAlgorithm")
    @click.option(
        "--imagestack", type=click.Path(exists=True))
    @click.pass_context
    def _cli(ctx, imagestack):
        ctx.obj["component"]._cli_run(
            ctx,
            AdditiveFilterAlgorithm(additive=ImageStack.from_path_or_url(imagestack)))


class FilterAlgorithmWithMissingConstructorTyping(FilterAlgorithmBase):
    def __init__(self, additive):
        self.additive = additive

    def run(self, stack: ImageStack, *args) -> ImageStack:
        numpy_array = stack.xarray
        numpy_array = numpy_array + stack.xarray
        return ImageStack.from_numpy(numpy_array)

    @staticmethod
    @click.command("FilterAlgorithmWithMissingConstructorTyping")
    @click.option(
        "--imagestack", type=click.Path(exists=True))
    @click.pass_context
    def _cli(ctx, imagestack):
        ctx.obj["component"]._cli_run(
            ctx,
            FilterAlgorithmWithMissingConstructorTyping(
                additive=ImageStack.from_path_or_url(imagestack)))


class FilterAlgorithmWithMissingRunTyping(FilterAlgorithmBase):
    def __init__(self, multiplicand: float):
        self.multiplicand = multiplicand

    def run(self, stack, *args) -> ImageStack:
        numpy_array = stack.xarray
        numpy_array = numpy_array * self.multiplicand
        return ImageStack.from_numpy(numpy_array)

    @staticmethod
    @click.command("FilterAlgorithmWithMissingRunTyping")
    @click.option(
        "--multiplicand", default=1.0, type=float)
    @click.pass_context
    def _cli(ctx, multiplicand):
        ctx.obj["component"]._cli_run(
            ctx, FilterAlgorithmWithMissingRunTyping(multiplicand=multiplicand))



from abc import abstractmethod
from typing import Type

from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.pipeline.pipelinecomponent import PipelineComponent
from starfish.core.util import click
from starfish.core.util.click.indirectparams import CodebookParamType


class Decode(PipelineComponent):
    """
    The Decode class exposes methods to compare detected spots to expected patterns of
    fluorescence across the rounds and channels of an experiment and map them to target genes or
    proteins.

    For single molecule FISH or RNAscope experiments, these codebooks are often simple mappings of
    (round, channel) pairs to targets. For coded assays, these codebooks can be much more complex.

    Example codebooks are associated with each experiment in :py:mod:`starfish.data` and can
    be accessed with :py:meth`Experiment.codebook`.
    """
    @classmethod
    def _cli_run(cls, ctx, instance):
        table = ctx.obj["intensities"]
        output = ctx.obj["output"]
        intensities: IntensityTable = instance.run(table)
        intensities.to_netcdf(output)

    @staticmethod
    @click.group("Decode")
    @click.option("-i", "--input", required=True, type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.option("--codebook", required=True, type=CodebookParamType)
    @click.pass_context
    def _cli(ctx, input, output, codebook):
        """assign genes to spots"""
        ctx.obj = dict(
            component=Decode,
            input=input,
            output=output,
            intensities=IntensityTable.open_netcdf(input),
            codebook=codebook,
        )


class DecodeAlgorithmBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return Decode

    @abstractmethod
    def run(self, intensities: IntensityTable, *args):
        """Performs decoding on the spots found, using the codebook specified."""
        raise NotImplementedError()



"""

Plotting Utilities for Documentation and Examples
=================================================

This module contains a series of utilities for creating two dimensional plots that are useful for
generating documentation and vignettes. We suggest that users leverage :py:func:`starfish.display`
for their plotting needs, as the interactive viewer is better able to handle the array of features
that starfish needs.
"""
import itertools
from typing import Any, Mapping, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.colors import ListedColormap

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Axes, Features


def imshow_plane(
    image_stack: ImageStack,
    sel: Optional[Mapping[Axes, Union[int, tuple]]] = None,
    ax=None,
    title: Optional[str] = None,
    **kwargs,
) -> None:
    """
    Plot a single plane of an ImageStack. If passed a selection function (sel), the stack will be
    subset using :py:meth:`ImageStack.sel`. If ax is passed, the function will be plotted in the
    provided axis. Additional kwargs are passed to :py:func:`plt.imshow`

    Parameters
    ----------
    image_stack : ImageStack
        imagestack from which to extract a 2-d image for plotting
    sel : Optional[Mapping[Axes, Union[int, tuple]]]
        Optional, but only if image_stack is already of shape (1, 1, 1, y, x). Selector to pass
        ImageStack.sel, Selects the (y, x) plane to be plotted.
    ax :
        Axes to plot on. If not passed, defaults to the current axes.
    title : Optional[str]
        Title to assign the Axes being plotted on.
    kwargs :
        additional keyword arguments to pass to plt.imshow

    """
    if ax is None:
        ax = plt.gca()

    if sel is not None:
        image_stack = image_stack.sel(sel)

    if title is not None:
        ax.set_title(title)

    # verify imagestack is 2d before trying to plot it
    data: xr.DataArray = image_stack.xarray.squeeze()
    if set(data.sizes.keys()).intersection({Axes.CH, Axes.ROUND, Axes.ZPLANE}):
        raise ValueError(f"image_stack must be a 2d (x, y) array, not {data.sizes}")

    # set imshow default kwargs
    if "cmap" not in kwargs:
        kwargs["cmap"] = plt.cm.gray

    ax.imshow(data, **kwargs)
    ax.axis("off")


def intensity_histogram(
    image_stack: ImageStack,
    sel: Optional[Mapping[Axes, Union[int, tuple]]] = None,
    ax=None,
    title: Optional[str] = None,
    **kwargs
) -> None:
    """
    Plot the 1-d intensity histogram of linearized image_stack.

    Parameters
    ----------
    image_stack : ImageStack
        imagestack containing intensities
    sel : Optional[Mapping[Axes, Union[int, tuple]]]
        Optional, Selector to pass ImageStack.sel that will restrict the histogram construction to
        the specified subset of image_stack.
    ax :
        Axes to plot on. If not passed, defaults to the current axes.
    title : Optional[str]
        Title to assign the Axes being plotted on.
    kwargs :
        additional keyword arguments to pass to plt.hist

    """
    if ax is None:
        ax = plt.gca()

    if sel is not None:
        image_stack = image_stack.sel(sel)

    if title is not None:
        ax.set_title(title)

    data: np.ndarray = np.ravel(image_stack.xarray)
    ax.hist(data, **kwargs)


def overlay_spot_calls(
    image_stack: ImageStack,
    intensities: IntensityTable,
    sel: Optional[Mapping[Axes, Union[int, tuple]]] = None,
    ax=None,
    title: Optional[str] = None,
    imshow_kwargs: Optional[Mapping[str, Any]] = None,
    scatter_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """
    Overlays spot calls atop a 2-d image extracted from ImageStack. Manages sub-selection from the
    IntensityTable and ImageStack based on provided `sel` parameter.

    Parameters
    ----------
    image_stack : ImageStack
        imagestack from which to extract a 2-d image for plotting
    intensities : IntensityTable
        contains spots to overlay on ImageStack.
    sel : Optional[Mapping[Axes, Union[int, tuple]]]
        Optional, but only if image_stack is already of shape (1, 1, 1, y, x). Selector to pass
        ImageStack.sel, Selects the (y, x) plane to be plotted. Will also be used to reduce the
        spots from intensities.
    ax :
        Axes to plot on. If not passed, defaults to the current axes.
    title : Optional[str]
        Title to assign the Axes being plotted on.
    imshow_kwargs : Optional[Mapping[str, Any]]
        additional keyword arguments to pass to imshow
    scatter_kwargs : Optional[Mapping[str, Any]]
        additional keyword arguments to pass to scatter
    """
    if ax is None:
        ax = plt.gca()

    if sel is not None:
        image_stack = image_stack.sel(sel)

        # subset the intensities if needed
        intensity_keys = (Axes.ROUND, Axes.CH, Axes.ZPLANE)
        intensity_sel = {x.value: sel[x] for x in intensity_keys if x in sel}
        if intensity_sel:
            intensities = intensities.sel(intensity_sel)

    imshow_kwargs = imshow_kwargs if imshow_kwargs else {}
    scatter_kwargs = scatter_kwargs if scatter_kwargs else {}

    # plot background
    imshow_plane(image_stack, sel=sel, ax=ax, title=title, **imshow_kwargs)

    # plot spots
    plt.scatter(
        x=np.asarray(intensities[Axes.X.value]),
        y=np.asarray(intensities[Axes.Y.value]),
        s=np.asarray(intensities[Features.SPOT_RADIUS]),
        c='red',
        **scatter_kwargs,
    )

    # reset the axes limits; scatter often extends them.
    ax.set_ylim((0, image_stack.shape[Axes.Y.value]))
    ax.set_xlim((0, image_stack.shape[Axes.X.value]))


def _linear_alpha_cmap(cmap):
    """add linear alpha to an existing colormap"""
    alpha_cmap = cmap(np.arange(cmap.N))
    alpha_cmap[:, -1] = np.linspace(0, 1, cmap.N)
    return ListedColormap(alpha_cmap)


def diagnose_registration(
    imagestack,
    *sels,
    ax=None,
    title: Optional[str] = None,
    **kwargs,
):
    """
    # TODO docme
    """
    if ax is None:
        ax = plt.gca()

    if title is not None:
        ax.set_title(title)

    # add linear alpha to existing colormaps to avoid the "white" color of the
    # map at 0 from clobbering previously plotted images. Functionally this
    # enables high intensity spots to be co-visualized in the same frame.
    cmaps = [
        plt.cm.Blues,
        plt.cm.Reds,
        plt.cm.Greens,
        plt.cm.Purples,
        plt.cm.Greys,
        plt.cm.Oranges
    ]

    alpha_cmap_cycle = itertools.cycle([_linear_alpha_cmap(cm) for cm in cmaps])

    for sel, cmap in zip(sels, alpha_cmap_cycle):
        imshow_plane(imagestack, sel, ax=ax, cmap=cmap)



import io
import tarfile
from typing import Dict, List, Sequence, Tuple, Union

import numpy as np
import xarray as xr
from skimage.measure import regionprops

from starfish.core.types import Axes, Coordinates


AXES = [a.value for a in Axes if a not in (Axes.ROUND, Axes.CH)]
COORDS = [c.value for c in Coordinates]


def _get_axes_names(ndim: int) -> Tuple[List[str], List[str]]:
    """Get needed axes names given the number of dimensions.

    Parameters
    ----------
    ndim : int
        Number of dimensions.

    Returns
    -------
    axes : List[str]
        Axes names.
    coords : List[str]
        Coordinates names.
    """
    if ndim == 2:
        axes = [axis for axis in AXES if axis != Axes.ZPLANE.value]
        coords = [coord for coord in COORDS if coord != Coordinates.Z.value]
    elif ndim == 3:
        axes = AXES
        coords = COORDS
    else:
        raise TypeError('expected 2- or 3-D image')

    return axes, coords


def validate_segmentation_mask(arr: xr.DataArray):
    """Validate if the given array is a segmentation mask.

    Parameters
    ----------
    arr : xr.DataArray
        Array to check.
    """
    if not isinstance(arr, xr.DataArray):
        raise TypeError(f"expected DataArray; got {type(arr)}")

    if arr.ndim not in (2, 3):
        raise TypeError(f"expected 2 or 3 dimensions; got {arr.ndim}")

    if arr.dtype != np.bool:
        raise TypeError(f"expected dtype of bool; got {arr.dtype}")

    axes, coords = _get_axes_names(arr.ndim)
    dims = set(axes)

    if dims != set(arr.dims):
        raise TypeError(f"missing dimensions '{dims.difference(arr.dims)}'")

    if dims.union(coords) != set(arr.coords):
        raise TypeError(f"missing coordinates '{dims.union(coords).difference(arr.coords)}'")


class SegmentationMaskCollection:
    """Collection of binary segmentation masks with a list-like access pattern.

    Parameters
    ----------
    masks : List[xr.DataArray]
        Segmentation masks.
    """
    def __init__(self, masks: List[xr.DataArray]):
        for mask in masks:
            validate_segmentation_mask(mask)

        self._masks = masks

    def __getitem__(self, index):
        return self._masks[index]

    def __iter__(self):
        return iter(self._masks)

    def __len__(self):
        return len(self._masks)

    def append(self, mask: xr.DataArray):
        """Add an existing segmentation mask.

        Parameters
        ----------
        arr : xr.DataArray
            Segmentation mask.
        """
        validate_segmentation_mask(mask)
        self._masks.append(mask)

    @classmethod
    def from_label_image(
            cls,
            label_image: np.ndarray,
            physical_ticks: Dict[Coordinates, Sequence[float]]
    ) -> "SegmentationMaskCollection":
        """Creates segmentation masks from a label image.

        Parameters
        ----------
        label_image : int array
            Integer array where each integer corresponds to a region.
        physical_ticks : Dict[Coordinates, Sequence[float]]
            Physical coordinates for each axis.

        Returns
        -------
        masks : SegmentationMaskCollection
            Masks generated from the label image.
        """
        props = regionprops(label_image)

        dims, _ = _get_axes_names(label_image.ndim)

        masks: List[xr.DataArray] = []

        coords: Dict[str, Union[list, Tuple[str, Sequence]]]

        # for each region (and its properties):
        for label, prop in enumerate(props):
            # create pixel coordinate labels from the bounding box
            # to preserve spatial indexing relative to the original image
            coords = {d: list(range(prop.bbox[i], prop.bbox[i + len(dims)]))
                      for i, d in enumerate(dims)}

            # create physical coordinate labels by taking the overlapping
            # subset from the full span of labels
            for d, c in physical_ticks.items():
                axis = d.value[0]
                i = dims.index(axis)
                coords[d.value] = (axis, c[prop.bbox[i]:prop.bbox[i + len(dims)]])

            name = str(label + 1)
            name = name.zfill(len(str(len(props))))  # pad with zeros

            mask = xr.DataArray(prop.image,
                                dims=dims,
                                coords=coords,
                                name=name)
            masks.append(mask)

        return cls(masks)

    @classmethod
    def from_disk(cls, path: str) -> "SegmentationMaskCollection":
        """Load the collection from disk.

        Parameters
        ----------
        path : str
            Path of the tar file to instantiate from.

        Returns
        -------
        masks : SegmentationMaskCollection
            Collection of segmentation masks.
        """
        masks = []

        with tarfile.open(path) as t:
            for info in t.getmembers():
                f = t.extractfile(info.name)
                mask = xr.open_dataarray(f)
                masks.append(mask)

        return cls(masks)

    def save(self, path: str):
        """Save the segmentation masks to disk.

        Parameters
        ----------
        path : str
            Path of the tar file to write to.
        """
        with tarfile.open(path, 'w:gz') as t:
            for i, mask in enumerate(self._masks):
                data = mask.to_netcdf()
                with io.BytesIO(data) as buff:
                    info = tarfile.TarInfo(name=str(i) + '.nc')
                    info.size = len(data)
                    t.addfile(tarinfo=info, fileobj=buff)



from abc import abstractmethod
from typing import Type

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.pipeline.pipelinecomponent import PipelineComponent
from starfish.core.util import click
from starfish.core.util.click.indirectparams import ImageStackParamType


class Filter(PipelineComponent):
    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        stack = ctx.obj["stack"]
        filtered = instance.run(stack)
        filtered.export(output)

    @staticmethod
    @click.group("Filter")
    @click.option("-i", "--input", type=ImageStackParamType)
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, input, output):
        """smooth, sharpen, denoise, etc"""
        print("Filtering images...")
        ctx.obj = dict(
            component=Filter,
            input=input,
            output=output,
            stack=input,
        )


class FilterAlgorithmBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return Filter

    @abstractmethod
    def run(self, stack: ImageStack, *args) -> ImageStack:
        """Perform filtering of an image stack"""
        raise NotImplementedError()



from typing import Optional, Tuple

import numpy as np

from starfish.core.codebook.codebook import Codebook
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.intensity_table.intensity_table_coordinates import \
    transfer_physical_coords_from_imagestack_to_intensity_table
from starfish.core.util import click
from ._base import DetectPixelsAlgorithmBase
from .combine_adjacent_features import CombineAdjacentFeatures, ConnectedComponentDecodingResult


class PixelSpotDecoder(DetectPixelsAlgorithmBase):
    """Decode an image by first coding each pixel, then combining the results into spots

    Parameters
    ----------
    codebook : Codebook
        Codebook object mapping codewords to the targets they are designed to detect
    metric : str
        the sklearn metric string to pass to NearestNeighbors
    distance_threshold : float
        spots whose codewords are more than this metric distance from an expected code are
        filtered
    magnitude_threshold : int
        spots with intensity less than this value are filtered
    min_area : int
        spots with total area less than this value are filtered
    max_area : int
        spots with total area greater than this value are filtered
    norm_order : int
        order of L_p norm to apply to intensities and codes when using metric_decode to pair
        each intensities to its closest target (default = 2)
    """
    def __init__(
            self, codebook: Codebook, metric: str, distance_threshold: float,
            magnitude_threshold: int, min_area: int, max_area: int, norm_order: int = 2
    ) -> None:

        self.codebook = codebook
        self.metric = metric
        self.distance_threshold = distance_threshold
        self.magnitude_threshold = magnitude_threshold
        self.min_area = min_area
        self.max_area = max_area
        self.norm_order = norm_order

    def run(
            self,
            primary_image: ImageStack,
            n_processes: Optional[int] = None,
            *args,
    ) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """decode pixels and combine them into spots using connected component labeling

        Parameters
        ----------
        primary_image : ImageStack
            ImageStack containing spots
        n_processes : Optional[int]
            The number of processes to use for CombineAdjacentFeatures.
             If None, uses the output of os.cpu_count() (default = None).

        Returns
        -------
        IntensityTable :
            IntensityTable containing decoded spots
        ConnectedComponentDecodingResult :
            Results of connected component labeling

        """
        pixel_intensities = IntensityTable.from_image_stack(primary_image)
        decoded_intensities = self.codebook.decode_metric(
            pixel_intensities,
            max_distance=self.distance_threshold,
            min_intensity=self.magnitude_threshold,
            norm_order=self.norm_order,
            metric=self.metric
        )
        caf = CombineAdjacentFeatures(
            min_area=self.min_area,
            max_area=self.max_area,
            mask_filtered_features=True
        )
        decoded_spots, image_decoding_results = caf.run(intensities=decoded_intensities,
                                                        n_processes=n_processes)

        transfer_physical_coords_from_imagestack_to_intensity_table(image_stack=primary_image,
                                                                    intensity_table=decoded_spots)
        return decoded_spots, image_decoding_results

    @staticmethod
    @click.command("PixelSpotDecoder")
    @click.option("--metric", type=str, default='euclidean')
    @click.option(
        "--distance-threshold", type=float, default=0.5176,
        help="maximum distance a pixel may be from a codeword before it is filtered"
    )
    @click.option(
        "--magnitude-threshold", type=float, default=1,
        help="minimum magnitude of a feature"
    )
    @click.option(
        "--min-area", type=int, default=2,
        help="minimum area of a feature"
    )
    @click.option(
        "--max-area", type=int, default=np.inf,
        help="maximum area of a feature"
    )
    @click.option(
        "--norm-order", type=int, default=2,
        help="order of L_p norm to apply to intensities "
        "and codes when using metric_decode to pair each intensities to its closest target"
    )
    @click.pass_context
    def _cli(
        ctx, metric, distance_threshold, magnitude_threshold, min_area, max_area, norm_order
    ):
        codebook = ctx.obj["codebook"]
        instance = PixelSpotDecoder(
            codebook=codebook,
            metric=metric,
            distance_threshold=distance_threshold,
            magnitude_threshold=magnitude_threshold,
            min_area=min_area,
            max_area=max_area,
            norm_order=norm_order,
        )
        ctx.obj["component"]._cli_run(ctx, instance)



from abc import abstractmethod
from typing import Type

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.pipeline import PipelineComponent
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.segmentation_mask import SegmentationMaskCollection
from starfish.core.util import click
from starfish.core.util.click.indirectparams import ImageStackParamType


class Segment(PipelineComponent):
    """
    Starfish class implementing segmentation approaches.
    """
    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        pri_stack = ctx.obj["primary_images"]
        nuc_stack = ctx.obj["nuclei"]

        masks = instance.run(pri_stack, nuc_stack)

        print(f"Writing masks to {output}")
        masks.save(output)

    @staticmethod
    @click.group("Segment")
    @click.option("--primary-images", required=True, type=ImageStackParamType)
    @click.option("--nuclei", required=True, type=ImageStackParamType)
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, primary_images, nuclei, output):
        """define masks for cell boundaries and assign spots"""
        print('Segmenting ...')
        ctx.obj = dict(
            component=Segment,
            output=output,
            primary_images=primary_images,
            nuclei=nuclei,
        )


class SegmentAlgorithmBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return Segment

    @abstractmethod
    def run(
            self,
            primary_image_stack: ImageStack,
            nuclei_stack: ImageStack,
            *args
    ) -> SegmentationMaskCollection:
        """Performs segmentation on the stack provided."""
        raise NotImplementedError()



from starfish.core.codebook.codebook import Codebook
from starfish.core.experiment.experiment import Experiment
from starfish.core.util.indirectfile._base import ConversionFormula


class GetCodebookFromExperiment(ConversionFormula):
    def applicable(self, input_parameter: str) -> bool:
        return input_parameter.startswith("@")

    def load(self, input_parameter: str) -> Codebook:
        path = input_parameter[1:]
        experiment = Experiment.from_json(path)
        return experiment.codebook


class GetCodebook(ConversionFormula):
    def applicable(self, input_parameter: str) -> bool:
        return not input_parameter.startswith("@")

    def load(self, input_parameter: str) -> Codebook:
        return Codebook.open_json(input_parameter)



from typing import Collection, Mapping, Sequence

import numpy as np

from starfish.core.types import Axes, Coordinates, Number
from ._key import TileKey


class TileData:
    """
    Base class for a parser to implement that provides the data for a single tile.
    """
    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        raise NotImplementedError()

    @property
    def numpy_array(self) -> np.ndarray:
        """Return the image data representing the tile.  The tile must be row-major.

        Returns
        -------
        ndarray :
            The image data
        """
        raise NotImplementedError()

    @property
    def coordinates(self) -> Mapping[Coordinates, Sequence[Number]]:
        raise NotImplementedError()

    @property
    def selector(self) -> Mapping[Axes, int]:
        raise NotImplementedError()


class TileCollectionData:
    """
    Base class for a parser to implement that provides the data for a collection of tiles to be
    assembled into an ImageStack.
    """
    def __getitem__(self, tilekey: TileKey) -> dict:
        """Returns the extras metadata for a given tile, addressed by its TileKey"""
        raise NotImplementedError()

    def keys(self) -> Collection[TileKey]:
        """Returns a Collection of the TileKey's for all the tiles."""
        raise NotImplementedError()

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        """Returns the shape of a tile."""
        raise NotImplementedError()

    @property
    def extras(self) -> dict:
        """Returns the extras metadata for the TileSet."""
        raise NotImplementedError()

    def get_tile_by_key(self, tilekey: TileKey) -> TileData:
        raise NotImplementedError()

    def get_tile(self, r: int, ch: int, z: int) -> TileData:
        raise NotImplementedError()



from time import time

import xarray as xr

from starfish.core.util import click
from . import validate_sptx


class DefaultGroup(click.Group):
    """
    Handle the old style invocation by changing --experiment to the subcommand
    """

    def parse_args(self, ctx, args):
        copy = list()
        for arg in args:
            if arg.startswith("--experiment-json"):
                parts = arg.split("=")
                parts[0] = "experiment"
                copy.extend(parts)
            else:
                copy.append(arg)
        super(DefaultGroup, self).parse_args(ctx, copy)


@click.group(cls=DefaultGroup)
@click.option("--experiment-json", required=False, metavar="JSON_FILE_OR_URL")
@click.option("--fuzz", is_flag=True)
@click.pass_context
def validate(ctx, experiment_json, fuzz):
    assert not experiment_json


@validate.command()
@click.argument("path", metavar="JSON_FILE_OR_URL")
@click.option("--fuzz", is_flag=True)
@click.pass_context
def codebook(ctx, path, fuzz):
    """validate codebook against the jsonschemas"""

    try:
        valid = validate_sptx.validate_file(path, "codebook/codebook.json", fuzz)
        if valid:
            ctx.exit(0)
        else:
            ctx.exit(1)
    except KeyboardInterrupt:
        ctx.exit(3)


@validate.command()
@click.argument("path", metavar="JSON_FILE_OR_URL")
@click.option("--fuzz", is_flag=True)
@click.pass_context
def experiment(ctx, path, fuzz):
    """validate experiment against the jsonschemas"""

    try:
        valid = validate_sptx.validate(path, fuzz)
        if valid:
            ctx.exit(0)
        else:
            ctx.exit(1)
    except KeyboardInterrupt:
        ctx.exit(3)


@validate.command()
@click.argument("path", metavar="JSON_FILE_OR_URL")
@click.option("--fuzz", is_flag=True)
@click.pass_context
def fov(ctx, path, fuzz):
    """validate field-of-view against the jsonschemas"""

    try:
        valid = validate_sptx.validate_file(path, "field_of_view/field_of_view.json", fuzz)
        if valid:
            ctx.exit(0)
        else:
            ctx.exit(1)
    except KeyboardInterrupt:
        ctx.exit(3)


@validate.command()
@click.argument("path", metavar="JSON_FILE_OR_URL")
@click.option("--fuzz", is_flag=True)
@click.pass_context
def manifest(ctx, path, fuzz):
    """validate manifest against the jsonschemas"""

    try:
        valid = validate_sptx.validate_file(path, "fov_manifest.json", fuzz)
        if valid:
            ctx.exit(0)
        else:
            ctx.exit(1)
    except KeyboardInterrupt:
        ctx.exit(3)


@validate.command()
@click.argument("path")
@click.pass_context
def xarray(ctx, path):
    try:
        d = xr.open_dataset(path)
        print("=" * 60)
        start = time()
        print(d)
        stop = time()
        print("=" * 60)
        print(f"Opened {path} in {stop-start}s.")
        names = set(d.coords._names)
        spots = "z y x radius z_min z_max y_min y_max x_min x_max "
        spots += "intensity spot_id features c r xc yc zc"
        spots = set(spots.split(" "))
        target_spots = set(["cell_id"])
        decoded_spots = set(["target", "distance", "passes_thresholds"])
        if spots.issubset(names):
            if target_spots.issubset(names):
                if decoded_spots.issubset(names):
                    print("Likely decoded spots")
                else:
                    print("Likely target spots")
            else:
                print("Likely just spots")
        else:
            print("Unknown xarray output format!")
            ctx.exit(1)
    except KeyboardInterrupt:
        ctx.exit(3)
    except Exception as e:
        print(f"Invalid xarray: {e}")
        ctx.exit(1)
    else:
        ctx.exit(0)



"""To build experiments in-place, a few things must be done:

1. Call enable_inplace_mode().
2. Call write_experiment_json with tile_opener=inplace_tile_opener.
3. The TileFetcher should return an instance of a InplaceTileFetcher.

Please note that enabling in-place experiment construction should be only done in an isolated script
dedicated to constructing an experiment, as it modifies some existing code paths.
"""


import abc
import io
import sys
from pathlib import Path
from typing import BinaryIO

import numpy as np
from slicedimage import Tile

from starfish.core.types import Axes
from .providers import FetchedTile


def sha256_get(tile_self):
    return tile_self.provider.sha256


def sha256_set(tile_self, value):
    pass


def enable_inplace_mode():
    Tile.sha256 = property(sha256_get, sha256_set)


def inplace_tile_opener(toc_path: Path, tile: Tile, file_ext: str) -> BinaryIO:
    return DevNull(tile.provider.filepath)


class DevNull(io.BytesIO):
    """A class meant to mimic an open(filepath, 'wb') operation but
    prevents any actual writing, reading, or seeking.
    This class is an ugly hack to prevent the slicedimage
    Writer.generate_partition_document() function from needlessly creating
    a copy of image data.
    See: https://docs.python.org/3/library/io.html
    See also: cpython/Lib/_pyio.py
    """
    def __init__(self, filepath: str, *args, **kwargs):
        super(DevNull, self).__init__(*args, **kwargs)
        self.name = filepath

    def read(self, size=-1):
        raise NotImplementedError()

    def write(self, b):
        return sys.getsizeof(b)

    def seek(self, pos, whence=0):
        raise NotImplementedError()


class InplaceFetchedTile(FetchedTile):
    """Data formatters that operate in-place should return tiles that extend this class."""
    @property
    @abc.abstractmethod
    def filepath(self) -> Path:
        """Returns the path of the source tile."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def sha256(self):
        """Returns the sha256 checksum of the source tile."""
        raise NotImplementedError()

    def tile_data(self) -> np.ndarray:
        return np.zeros(shape=(self.shape[Axes.Y], self.shape[Axes.X]), dtype=np.float32)



import os
import warnings
from json import dump, loads

from diskcache import Cache
from pytest import mark, raises

from starfish import data
from starfish.core.util.config import Config, NestedDict
from .. import environ, StarfishConfig


def test_nested_dict():
    rd = dict()
    nd = NestedDict()
    rd[1] = {2: None}
    nd[3][4][5][6] = 7
    nd.update(rd)
    nd[1][2] = 9
    nd[1][8] = 10


simple_str = '{"a": 1}'
simple_map = loads(simple_str)

deep_str = '{"a": {"b": {"c": [1, 2, 3]}}}'
deep_map = loads(deep_str)


def test_simple_config_value_str():
    config = Config(simple_str)
    assert config.data["a"] == 1


def test_simple_config_value_map():
    config = Config(simple_map)
    assert config.data["a"] == 1


def test_simple_config_false():
    # config says the value is False.  Do we return False?
    config = Config("{\"a\": false}")
    assert config.lookup(("a",), None) is False

    # config says the value is not known.  Do we return the default valueS?
    config = Config("{}")
    assert config.lookup(("a",), None) is None
    assert config.lookup(("a",), False) is False


def test_simple_config_value_file(tmpdir):
    f = tmpdir.join("config.json")
    f.write(simple_str)
    config = Config(f"@{f}")
    assert config.data["a"] == 1


def test_lookup_dne():
    config = Config(simple_str)
    with raises(KeyError):
        config.lookup(["foo"])
    assert config.lookup(["foo"], 1) == 1
    assert config.lookup(["foo", "bar"], 2) == 2


def test_lookup_deep():
    config = Config(deep_str)
    assert config.lookup(["a"]) == {"b": {"c": [1, 2, 3]}}
    assert config.lookup(["a", "b"]) == {"c": [1, 2, 3]}
    assert config.lookup(["a", "b", "c"]) == [1, 2, 3]
    with raises(AttributeError):
        config.lookup(["a", "b", "c", "d"])
    assert config.lookup(["a", "b", "c", "d"], "x") == "x"


def test_cache_config():
    config = Config("""{
        "caching": {
             "enabled": true,
             "size_limit": 5e9,
             "directory": "/tmp"
         }
    }""")
    cache_config = config.lookup(("caching",), {})
    assert cache_config["enabled"]
    assert cache_config["size_limit"] == 5 * 10 ** 9


def test_cache_remove_step_wise():
    config = Config("""{
        "a": {
             "b": {
                "c": true,
                "d": true
             },
             "e": 1
         },
         "f": 2
    }""")
    config.lookup(("a", "b", "c"), {}, remove=True)
    assert "c" not in config.data["a"]["b"]
    config.lookup(("a", "b"), {}, remove=True)
    assert "b" not in config.data["a"]
    config.lookup(("a",), {}, remove=True)
    assert "a" not in config.data


def test_cache_remove_all():
    config = Config("""{
        "a": {
             "b": {
                "c": true
             }
         }
    }""")
    config.lookup(("a", "b", "c"), {}, remove=True)
    assert not config.data


def test_starfish_config_value_default_key(monkeypatch):
    with monkeypatch.context() as mc:
        mc.setenv("STARFISH_CONFIG", simple_str)
        config = StarfishConfig()._config_obj
        assert config.data["a"] == 1


@mark.parametrize("name,expected,config", (
    ("enabled", (2658848, 4e6), {
        "validation": {"strict": True},
        "slicedimage": {
            "caching": {
                "directory": "REPLACEME",
            }}}),
    ("disabled", (0, 0), {
        "validation": {"strict": True},
        "slicedimage": {
            "caching": {
                "size_limit": 0,
            }}}),
    ("limited", (1e5, 3e6), {
        "validation": {"strict": True},
        "slicedimage": {
            "caching": {
                "directory": "REPLACEME",
                "size_limit": 1e5,
            }}}),
))
def test_cache_merfish(tmpdir, name, expected, config, monkeypatch):

    cache_enabled = (0 != config["slicedimage"]["caching"].get("size_limit", None))
    if cache_enabled:
        config["slicedimage"]["caching"]["directory"] = str(tmpdir / "caching")

    with monkeypatch.context() as mc:
        setup_config(config, tmpdir, mc)

        # Run 1
        data.MERFISH(use_test_data=True).fov().get_image("primary")

        # Run 2
        if cache_enabled:
            data.MERFISH(use_test_data=True).fov().get_image("primary")

        # Check constraints
        if cache_enabled:
            # Enforce smallest size
            cache = Cache(str(tmpdir / "caching"))
            cache.cull()

        cache_size = get_size(tmpdir / "caching")
        min, max = expected
        assert (min <= cache_size) and (cache_size <= max)


def test_starfish_config(tmpdir, monkeypatch):
    with monkeypatch.context() as mc:
        config = {"slicedimage": {"caching": {"size_limit": 0}}}
        setup_config(config, tmpdir, mc, STARFISH_SLICEDIMAGE_CACHING_SIZE_LIMIT="1")
        assert 1 == StarfishConfig().slicedimage["caching"]["size_limit"]

    with monkeypatch.context() as mc:
        config = {"slicedimage": {"caching": {"size_limit": 0}}}
        setup_config(config, tmpdir, mc, SLICEDIMAGE_CACHING_SIZE_LIMIT="1")
        assert 1 == StarfishConfig().slicedimage["caching"]["size_limit"]

    with monkeypatch.context() as mc:
        config = {"validation": {"strict": True}}
        setup_config(config, tmpdir, mc)
        assert StarfishConfig().strict

    with monkeypatch.context() as mc:
        config = {"validation": {"strict": False}}
        setup_config(config, tmpdir, mc, STARFISH_VALIDATION_STRICT=None)  # Disable from travis
        assert not StarfishConfig().strict

    with monkeypatch.context() as mc:
        setup_config({}, tmpdir, mc, STARFISH_VALIDATION_STRICT="true")
        assert StarfishConfig().strict

    with monkeypatch.context() as mc:
        setup_config({}, tmpdir, mc, STARFISH_VALIDATION_STRICT="false")
        assert not StarfishConfig().strict


def test_starfish_warn(tmpdir, monkeypatch):
    config = {"unknown": True}
    with monkeypatch.context() as mc:
        setup_config(config, tmpdir, mc, STARFISH_SLICEDIMAGE_CACHING_SIZE_LIMIT="1")
        with warnings.catch_warnings(record=True) as warnings_:
            StarfishConfig()
            assert len(warnings_) == 1  # type: ignore


def test_starfish_environ(monkeypatch):
    with monkeypatch.context() as mc:
        mc.setenv("STARFISH_CONFIG", "{}")
        assert not StarfishConfig().strict
        with environ(VALIDATION_STRICT="true"):
            assert StarfishConfig().strict


def test_starfish_environ_warn(tmpdir, monkeypatch):
    with monkeypatch.context() as mc:
        setup_config({}, tmpdir, mc)
        with environ(UNKNOWN="true"):
            with warnings.catch_warnings(record=True) as warnings_:
                StarfishConfig()
                assert len(warnings_) == 1  # type: ignore


def test_starfish_environ_warn_on_duplicate(tmpdir, monkeypatch):
    with monkeypatch.context() as mc:
        setup_config({}, tmpdir, mc,
                     SLICEDIMAGE_CACHING_DIRECTORY=None,
                     STARFISH_SLICEDIMAGE_CACHING_DIRECTORY=None)
        assert "directory" not in StarfishConfig().slicedimage
        with environ(SLICEDIMAGE_CACHING_DIRECTORY="foo",
                     STARFISH_SLICEDIMAGE_CACHING_DIRECTORY="bar"):
            with warnings.catch_warnings(record=True) as warnings_:
                assert "bar" == StarfishConfig().slicedimage["caching"]["directory"]
                assert len(warnings_) == 1  # type: ignore

#
# HELPERS
#


def get_size(start_path='.'):
    """helper method for listing file sizes in a directory"""
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def setup_config(config, tmpdir, monkeypatch, **environment_variables):
    config_file = tmpdir / "config"
    with open(config_file, "w") as o:
        dump(config, o)
    monkeypatch.setitem(os.environ, "STARFISH_CONFIG", f"@{config_file}")
    for k, v in environment_variables.items():
        if v is None:
            monkeypatch.delitem(os.environ, k, raising=False)
        else:
            monkeypatch.setitem(os.environ, k, v)



import numpy as np
import pytest

from starfish import ImageStack
from starfish.core.imagestack.test.factories import unique_tiles_imagestack
from starfish.core.test.factories import (
    two_spot_informative_blank_coded_data_factory,
    two_spot_one_hot_coded_data_factory,
    two_spot_sparse_coded_data_factory,
)
from starfish.types import Axes, Features
from .._base import DetectSpotsAlgorithmBase
from ..blob import BlobDetector
from ..detect import detect_spots
from ..local_max_peak_finder import LocalMaxPeakFinder
from ..trackpy_local_max_peak_finder import TrackpyLocalMaxPeakFinder

# verify all spot finders handle different coding types
_, ONE_HOT_IMAGESTACK, ONE_HOT_MAX_INTENSITY = two_spot_one_hot_coded_data_factory()
_, SPARSE_IMAGESTACK, SPARSE_MAX_INTENSITY = two_spot_sparse_coded_data_factory()
_, BLANK_IMAGESTACK, BLANK_MAX_INTENSITY = two_spot_informative_blank_coded_data_factory()

# make sure that all spot finders handle empty arrays
EMPTY_IMAGESTACK = ImageStack.from_numpy(np.zeros((4, 2, 10, 100, 100), dtype=np.float32))


def simple_gaussian_spot_detector() -> BlobDetector:
    """create a basic gaussian spot detector"""
    return BlobDetector(min_sigma=1, max_sigma=4, num_sigma=5, threshold=0, measurement_type='max')


def simple_trackpy_local_max_spot_detector() -> TrackpyLocalMaxPeakFinder:
    """create a basic local max peak finder"""
    return TrackpyLocalMaxPeakFinder(
        spot_diameter=3,
        min_mass=0.01,
        max_size=10,
        separation=2,
    )


def simple_local_max_spot_detector() -> LocalMaxPeakFinder:
    return LocalMaxPeakFinder(
        min_distance=6,
        stringency=0,
        min_obj_area=0,
        max_obj_area=np.inf,
        threshold=0
    )


# initialize spot detectors
gaussian_spot_detector = simple_gaussian_spot_detector()
trackpy_local_max_spot_detector = simple_trackpy_local_max_spot_detector()
local_max_spot_detector = simple_local_max_spot_detector()

# test parameterization
test_parameters = (
    'data_stack, spot_detector, radius_is_gyration, max_intensity',
    [
        (ONE_HOT_IMAGESTACK, gaussian_spot_detector, False, ONE_HOT_MAX_INTENSITY),
        (ONE_HOT_IMAGESTACK, trackpy_local_max_spot_detector, True, ONE_HOT_MAX_INTENSITY),
        (ONE_HOT_IMAGESTACK, local_max_spot_detector, False, ONE_HOT_MAX_INTENSITY),
        (SPARSE_IMAGESTACK, gaussian_spot_detector, False, SPARSE_MAX_INTENSITY),
        (SPARSE_IMAGESTACK, trackpy_local_max_spot_detector, True, SPARSE_MAX_INTENSITY),
        (SPARSE_IMAGESTACK, local_max_spot_detector, False, SPARSE_MAX_INTENSITY),
        (BLANK_IMAGESTACK, gaussian_spot_detector, False, BLANK_MAX_INTENSITY),
        (BLANK_IMAGESTACK, trackpy_local_max_spot_detector, True, BLANK_MAX_INTENSITY),
        (BLANK_IMAGESTACK, local_max_spot_detector, False, BLANK_MAX_INTENSITY),
    ]
)


@pytest.mark.parametrize(*test_parameters)
def test_spot_detection_with_reference_image(
        data_stack: ImageStack,
        spot_detector: DetectSpotsAlgorithmBase,
        radius_is_gyration: bool,
        max_intensity: float,
):
    """This testing method uses a reference image to identify spot locations."""

    def call_detect_spots(stack):

        return detect_spots(
            data_stack=stack,
            spot_finding_method=spot_detector.image_to_spots,
            reference_image=stack,
            reference_image_max_projection_axes=(Axes.ROUND, Axes.CH),
            measurement_function=np.max,
            radius_is_gyration=radius_is_gyration,
            n_processes=1,
        )

    intensity_table = call_detect_spots(data_stack)
    assert intensity_table.sizes[Features.AXIS] == 2, "wrong number of spots detected"
    expected = [max_intensity * 2, max_intensity * 2]
    assert np.allclose(intensity_table.sum((Axes.ROUND, Axes.CH)).values, expected), \
        "wrong spot intensities detected"

    # verify this execution strategy produces an empty intensitytable when called with a blank image
    empty_intensity_table = call_detect_spots(EMPTY_IMAGESTACK)
    assert empty_intensity_table.sizes[Features.AXIS] == 0


@pytest.mark.parametrize(*test_parameters)
def test_spot_detection_with_reference_image_from_max_projection(
        data_stack: ImageStack,
        spot_detector: DetectSpotsAlgorithmBase,
        radius_is_gyration: bool,
        max_intensity: float,
):
    """This testing method builds a reference image to identify spot locations."""

    def call_detect_spots(stack):
        return detect_spots(
            data_stack=stack,
            spot_finding_method=spot_detector.image_to_spots,
            reference_image=stack,
            reference_image_max_projection_axes=(Axes.ROUND, Axes.CH),
            measurement_function=np.max,
            radius_is_gyration=radius_is_gyration,
            n_processes=1,
        )

    intensity_table = call_detect_spots(data_stack)
    assert intensity_table.sizes[Features.AXIS] == 2, "wrong number of spots detected"
    expected = [max_intensity * 2, max_intensity * 2]
    assert np.allclose(intensity_table.sum((Axes.ROUND, Axes.CH)).values, expected), \
        "wrong spot intensities detected"

    empty_intensity_table = call_detect_spots(EMPTY_IMAGESTACK)
    assert empty_intensity_table.sizes[Features.AXIS] == 0

@pytest.mark.parametrize(*test_parameters)
def test_spot_finding_no_reference_image(
        data_stack: ImageStack,
        spot_detector: DetectSpotsAlgorithmBase,
        radius_is_gyration: bool,
        max_intensity: float,
):
    """
    This testing method does not provide a reference image, and should therefore check for spots
    in each (round, ch) combination in sequence. With the given input, it should detect 4 spots.
    """

    def call_detect_spots(stack):
        return detect_spots(
            data_stack=stack,
            spot_finding_method=spot_detector.image_to_spots,
            measurement_function=np.max,
            radius_is_gyration=radius_is_gyration,
            n_processes=1,
        )

    intensity_table = call_detect_spots(data_stack)
    assert intensity_table.sizes[Features.AXIS] == 4, "wrong number of spots detected"
    expected = [max_intensity] * 4
    assert np.allclose(intensity_table.sum((Axes.ROUND, Axes.CH)).values, expected), \
        "wrong spot intensities detected"

    empty_intensity_table = call_detect_spots(EMPTY_IMAGESTACK)
    assert empty_intensity_table.sizes[Features.AXIS] == 0


def _make_labeled_image() -> ImageStack:
    ROUND_LABELS = (1, 4, 6)
    CH_LABELS = (2, 4, 6, 8)
    ZPLANE_LABELS = (3, 4)
    HEIGHT = 2
    WIDTH = 4

    return unique_tiles_imagestack(
        ROUND_LABELS, CH_LABELS, ZPLANE_LABELS, HEIGHT, WIDTH)


def test_reference_image_spot_detection_with_image_with_labeled_axes(monkeypatch):
    """This testing method uses a reference image to identify spot locations."""

    def call_detect_spots(stack):

        return detect_spots(
            data_stack=stack,
            spot_finding_method=gaussian_spot_detector.image_to_spots,
            reference_image=stack,
            reference_image_max_projection_axes=(Axes.ROUND, Axes.CH),
            measurement_function=np.max,
            radius_is_gyration=False,
            n_processes=1,
        )

    data_stack = _make_labeled_image()
    call_detect_spots(data_stack)


def test_spot_detection_with_image_with_labeled_axes():
    """This testing method uses no reference image to identify spot locations."""

    def call_detect_spots(stack):

        return detect_spots(
            data_stack=stack,
            spot_finding_method=gaussian_spot_detector.image_to_spots,
            measurement_function=np.max,
            radius_is_gyration=False,
            n_processes=1,
        )

    data_stack = _make_labeled_image()
    call_detect_spots(data_stack)



import os
import tempfile
import warnings

import numpy as np
import pytest

from starfish import ImageStack
from . import fakefilter  # noqa: F401
from ..errors import ConstructorError, ConstructorExtraParameterWarning
from ..recipe import Recipe


BASE_EXPECTED = np.array([
    [0.227543, 0.223117, 0.217014, 0.221241, 0.212863, 0.211963, 0.210575,
     0.198611, 0.194827, 0.181964],
    [0.216617, 0.214710, 0.212467, 0.218158, 0.211429, 0.210361, 0.205737,
     0.190814, 0.182010, 0.165667],
    [0.206744, 0.204685, 0.208774, 0.212909, 0.215274, 0.206180, 0.196674,
     0.179080, 0.169207, 0.157549],
    [0.190845, 0.197131, 0.188540, 0.195361, 0.196765, 0.200153, 0.183627,
     0.167590, 0.159930, 0.150805],
    [0.181231, 0.187457, 0.182910, 0.179416, 0.175357, 0.172137, 0.165072,
     0.156344, 0.153735, 0.150378],
    [0.169924, 0.184604, 0.182422, 0.174441, 0.159823, 0.157229, 0.157259,
     0.151690, 0.147265, 0.139940],
    [0.164874, 0.169467, 0.178012, 0.173129, 0.161425, 0.155978, 0.152712,
     0.150286, 0.145159, 0.140658],
    [0.164508, 0.165042, 0.171420, 0.174990, 0.162951, 0.152422, 0.149325,
     0.151675, 0.141588, 0.139010],
    [0.162448, 0.156451, 0.158419, 0.162722, 0.160388, 0.152865, 0.142885,
     0.142123, 0.140093, 0.135836],
    [0.150072, 0.147295, 0.145495, 0.153216, 0.156085, 0.149981, 0.145571,
     0.141878, 0.138857, 0.136965]],
    dtype=np.float32)
URL = "https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/fov_001/hybridization.json"


def test_simple_recipe():
    """Test that a simple recipe can execute correctly."""
    recipe_str = """
file_outputs[0] = compute(Filter.SimpleFilterAlgorithm, file_inputs[0], multiplicand=0.5)
    """

    with tempfile.TemporaryDirectory() as tempdir:
        output_path = os.path.join(tempdir, "output.json")
        recipe = Recipe(recipe_str, [URL], [output_path])

        execution = recipe._execution()
        execution.run_and_save()

        result_stack = ImageStack.from_path_or_url(output_path)
        assert np.allclose(
            BASE_EXPECTED * .5,
            result_stack.xarray[2, 2, 0, 40:50, 40:50]
        )


def test_chained_recipe():
    """Test that a recipe with a complex graph can execute correctly."""
    recipe_str = """
a = compute(Filter.SimpleFilterAlgorithm, file_inputs[0], multiplicand=0.5)
b = compute(Filter.SimpleFilterAlgorithm, a, multiplicand=.3)
file_outputs[0] = compute(Filter.SimpleFilterAlgorithm, b, multiplicand=0.2)
c = compute(Filter.SimpleFilterAlgorithm, a, multiplicand=.2)
file_outputs[1] = compute(Filter.SimpleFilterAlgorithm, c, multiplicand=.3)
    """

    with tempfile.TemporaryDirectory() as tempdir:
        output_0_path = os.path.join(tempdir, "output_0.json")
        output_1_path = os.path.join(tempdir, "output_1.json")
        recipe = Recipe(recipe_str, [URL], [output_0_path, output_1_path])

        execution = recipe._execution()
        execution.run_and_save()

        for path in (output_0_path, output_1_path):
            result_stack = ImageStack.from_path_or_url(path)
            assert np.allclose(
                BASE_EXPECTED * .03,
                result_stack.xarray[2, 2, 0, 40:50, 40:50]
            )


def test_garbage_collection():
    """Test that recipe execution discards intermediate results that are no longer necessary."""
    recipe_str = """
a = compute(Filter.SimpleFilterAlgorithm, file_inputs[0], multiplicand=0.5)
b = compute(Filter.SimpleFilterAlgorithm, a, multiplicand=0.3)
c = compute(Filter.SimpleFilterAlgorithm, b, multiplicand=4.)
d = compute(Filter.SimpleFilterAlgorithm, c, multiplicand=0.5)
file_outputs[0] = compute(Filter.SimpleFilterAlgorithm, d, multiplicand=0.5)
    """

    with tempfile.TemporaryDirectory() as tempdir:
        output_path = os.path.join(tempdir, "output.json")
        recipe = Recipe(recipe_str, [URL], [output_path])

        execution = recipe._execution()
        execution._run_one_tick()
        execution._run_one_tick()

        assert len(execution._completed_results) == 1

        execution.run_and_save()

        result_stack = ImageStack.from_path_or_url(output_path)
        assert np.allclose(
            BASE_EXPECTED * .15,
            result_stack.xarray[2, 2, 0, 40:50, 40:50]
        )


def test_recipe_constructor_missing_args():
    """Test that recipe construction detects missing arguments to the constructor."""
    recipe_str = """
file_output[0] = compute(Filter.SimpleFilterAlgorithm, file_inputs[0])
"""

    with tempfile.TemporaryDirectory() as tempdir:
        output_path = os.path.join(tempdir, "output.json")
        with pytest.raises(ConstructorError):
            Recipe(recipe_str, [URL], [output_path])


def test_recipe_constructor_extra_args():
    """Test that recipe construction detects missing arguments to the constructor."""
    recipe_str = """
file_outputs[0] = compute(Filter.SimpleFilterAlgorithm, file_inputs[0], multiplicand=.5, x=1)
"""

    with tempfile.TemporaryDirectory() as tempdir:
        output_path = os.path.join(tempdir, "output.json")
        with warnings.catch_warnings(record=True) as w:
            recipe = Recipe(recipe_str, [URL], [output_path])

            assert len(w) == 1
            assert issubclass(w[-1].category, ConstructorExtraParameterWarning)

        execution = recipe._execution()
        execution.run_and_save()

        result_stack = ImageStack.from_path_or_url(output_path)
        assert np.allclose(
            BASE_EXPECTED * .5,
            result_stack.xarray[2, 2, 0, 40:50, 40:50]
        )



from typing import Callable, Collection, Mapping, Optional, Sequence, Tuple, Type

from starfish.core.experiment.builder import TileIdentifier
from starfish.core.experiment.builder.test.factories.all_purpose import (
    collection_factory,
    LocationAwareFetchedTile,
)
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.imagestack.parser.crop import CropParameters
from starfish.core.types import Coordinates, CoordinateValue, Number


def imagestack_factory(
        fetched_tile_cls: Type[LocationAwareFetchedTile],
        round_labels: Sequence[int],
        ch_labels: Sequence[int],
        zplane_labels: Sequence[int],
        tile_height: int,
        tile_width: int,
        xrange: Tuple[Number, Number],
        yrange: Tuple[Number, Number],
        zrange: Tuple[Number, Number],
        crop_parameters: Optional[CropParameters] = None) -> ImageStack:
    """Given a type that implements the :py:class:`LocationAwareFetchedTile` contract, produce an
    imagestack with those tiles, and apply coordinates such that the 5D tensor has coordinates
    that range from `xrange[0]:xrange[1]`, `yrange[0]:yrange[1]`, `zrange[0]:zrange[1]`.

    Parameters
    ----------
    fetched_tile_cls : Type[LocationAwareFetchedTile]
        The class of the FetchedTile.
    round_labels : Sequence[int]
        Labels for the rounds.
    ch_labels : Sequence[int]
        Labels for the channels.
    zplane_labels : Sequence[int]
        Labels for the zplanes.
    tile_height : int
        Height of each tile, in pixels.
    tile_width : int
        Width of each tile, in pixels.
    xrange : Tuple[Number, Number]
        The starting and ending x physical coordinates for the tile.
    yrange : Tuple[Number, Number]
        The starting and ending y physical coordinates for the tile.
    zrange : Tuple[Number, Number]
        The starting and ending z physical coordinates for the tile.
    crop_parameters : Optional[CropParameters]
        The crop parameters to apply during ImageStack construction.
    """
    tile_identifiers: Collection[TileIdentifier] = [
        TileIdentifier(0, round_label, ch_label, zplane_label)
        for round_label in round_labels
        for ch_label in ch_labels
        for zplane_label in zplane_labels
    ]

    def make_tile_coordinate_callback(
            all_zplane_labels: Sequence[int]
    ) -> Callable[[TileIdentifier], Mapping[Coordinates, CoordinateValue]]:
        def tile_coordinate_callback(
                tile_identifier: TileIdentifier
        ) -> Mapping[Coordinates, CoordinateValue]:
            zplane_offset = all_zplane_labels.index(tile_identifier.zplane_label)
            return {
                Coordinates.X: xrange,
                Coordinates.Y: yrange,
                Coordinates.Z: zrange[zplane_offset],
            }

        return tile_coordinate_callback

    collection = collection_factory(
        fetched_tile_cls,
        tile_identifiers,
        make_tile_coordinate_callback(
            sorted(set(tile_identifier.zplane_label for tile_identifier in tile_identifiers))),
        tile_height,
        tile_width,
    )
    tileset = list(collection.all_tilesets())[0][1]

    return ImageStack.from_tileset(tileset, crop_parameters)



class RecipeWarning(RuntimeWarning):
    pass


class RecipeError(Exception):
    pass


class ConstructorExtraParameterWarning(RecipeWarning):
    """Raised when a recipe contains parameters that an algorithms constructor does not expect."""


class TypeInferenceError(RecipeError):
    """Raised when we cannot infer the type of object an algorithm expects in its constructor or
    its run method.  This can be fixed by ensuring all the parameters to the constructor and the run
    method have type hints."""


class ConstructorError(RecipeError):
    """Raised when there is an error raised during the construction of an algorithm class."""
    pass


class RunInsufficientParametersError(RecipeError):
    """Raised when the recipe does not provide sufficient parameters for the run method."""


class ExecutionError(RecipeError):
    """Raised when there is an error raised during the execution of an algorithm."""
    pass



import os
import shutil
from pathlib import Path
from typing import Iterable, Optional

from starfish.core.util import exec


class RecipeTest:
    """This is a base class for testing recipes.  Each recipe test should define its recipe file,
    the input files, the output files, and a test method that verifies the correctness of the
    results.
    """
    @property
    def recipe(self) -> Path:
        raise NotImplementedError()

    @property
    def input_url_or_paths(self) -> Iterable[str]:
        raise NotImplementedError()

    @property
    def output_paths(self) -> Iterable[Path]:
        raise NotImplementedError()

    def verify_results(self, tempdir: Path):
        raise NotImplementedError()

    def test_run_recipe(self):
        cmdline = ["starfish", "recipe", "--recipe", self.recipe]
        for input_url_or_path in self.input_url_or_paths:
            cmdline.extend(["--input", input_url_or_path])
        for output_path in self.output_paths:
            cmdline.extend([
                "--output",
                lambda tempdir, *args, **kwargs: os.path.join(tempdir, os.fspath(output_path))])

        tempdir: Optional[str] = None
        try:
            tempdir = exec.stages([cmdline], keep_data=True)

            self.verify_results(Path(tempdir))
        finally:
            if tempdir is not None and os.getenv("TEST_KEEP_DATA") is None:
                shutil.rmtree(tempdir)



from .errors import (
    ConstructorError,
    ConstructorExtraParameterWarning,
    ExecutionError,
    RecipeError,
    RunInsufficientParametersError,
    TypeInferenceError,
)
from .recipe import Recipe
from .runnable import Runnable



import numpy as np
import xarray as xr

from starfish import IntensityTable
from starfish.core.codebook.test.factories import codebook_array_factory, loaded_codebook
from starfish.core.types import Coordinates, Features
from ..overlap import Area


def synthetic_intensity_table() -> IntensityTable:
    return IntensityTable.synthetic_intensities(loaded_codebook(), n_spots=2)


def create_intensity_table_with_coords(area: Area, n_spots: int=10) -> IntensityTable:
    """
    Creates a 50X50 intensity table with physical coordinates within
    the given Area.

    Parameters
    ----------
    area: Area
        The area of physical space the IntensityTable should be defined over
    n_spots:
        Number of spots to add to the IntensityTable
    """
    codebook = codebook_array_factory()
    it = IntensityTable.synthetic_intensities(
        codebook,
        num_z=1,
        height=50,
        width=50,
        n_spots=n_spots
    )
    # intensity table 1 has 10 spots, xmin = 0, ymin = 0, xmax = 2, ymax = 1
    it[Coordinates.X.value] = xr.DataArray(np.linspace(area.min_x, area.max_x, n_spots),
                                           dims=Features.AXIS)
    it[Coordinates.Y.value] = xr.DataArray(np.linspace(area.min_y, area.max_y, n_spots),
                                           dims=Features.AXIS)
    return it



import importlib
from abc import abstractmethod
from pathlib import Path
from typing import Mapping, Optional, Set, Type


class PipelineComponent:
    """
    This is the base class for any method executed by the CLI.

    PipelineComponent is an Abstract Class that exposes two private methods to link any subclassing
    method to the CLI, _algorithm_to_class_map, which fetches all the algorithms that extend this
    base class at run time, and _cli_register, which registers those methods to the CLI. It exposes
    two additional abstract private methods that must be extended by subclasses:

    Methods
    -------
    _get_algorithm_base_class()
        should simply return an instance of the AlgorithmBase. See, e.g.
        starfish.image.segmentation.Segmentation
    _cli_run(ctx, instance, *args, **kwargs)
        implements the behavior of the pipeline component that must occur when this component is
        evoked from the CLI. This often includes loading serialized objects into memory and
        passing them to the API's run command.

    See Also
    --------
    starfish.pipeline.algorithmbase.py
    """

    _algorithm_to_class_map_int: Optional[Mapping[str, Type]] = None

    @classmethod
    def _algorithm_to_class_map(cls) -> Mapping[str, Type]:
        """Returns a mapping from algorithm names to the classes that implement them."""
        assert cls._algorithm_to_class_map_int is not None
        return cls._algorithm_to_class_map_int

    @classmethod
    @abstractmethod
    def _cli_run(cls, ctx, instance):
        raise NotImplementedError()


def import_all_submodules(path_str: str, package: str, excluded: Optional[Set[str]]=None) -> None:
    """
    Given a path of a __init__.py file, find all the .py files in that directory and import them
    relatively to a package.

    Parameters
    ----------
    path_str : str
        The path of a __init__.py file.
    package : str
        The package name that the modules should be imported relative to.
    excluded : Optional[Set[str]]
        A set of files not to include.  If this is not provided, it defaults to set("__init__.py").
    """
    if excluded is None:
        excluded = {"__init__.py"}

    path: Path = Path(path_str).parent
    for entry in path.iterdir():
        if not entry.suffix.lower().endswith(".py"):
            continue
        if entry.name.lower() in excluded:
            continue

        importlib.import_module(f".{entry.stem}", package)



"""
These tests center around creating an ImageStack but selectively loading data from the original
TileSet.
"""
import numpy as np

from starfish.core.experiment.builder.test.factories.unique_tiles import unique_data
from starfish.core.types import Axes
from .factories.unique_tiles import (
    unique_tiles_imagestack, X_COORDS, Y_COORDS, Z_COORDS,
)
from .imagestack_test_utils import (
    recalculate_physical_coordinate_range,
    verify_physical_coordinates,
    verify_stack_data,
)
from ..imagestack import ImageStack
from ..parser.crop import CropParameters


NUM_FOV = 1
NUM_ROUND = 3
NUM_CH = 4
NUM_ZPLANE = 2

ROUND_LABELS = list(range(NUM_ROUND))
CH_LABELS = list(range(NUM_CH))
ZPLANE_LABELS = list(range(NUM_ZPLANE))
HEIGHT = 40
WIDTH = 60


def expected_data(round_label: int, ch_label: int, zplane_label: int):
    return unique_data(
        0, round_label, ch_label, zplane_label,
        NUM_FOV, NUM_ROUND, NUM_CH, NUM_ZPLANE,
        HEIGHT, WIDTH
    )


def setup_imagestack(crop_parameters: CropParameters) -> ImageStack:
    return unique_tiles_imagestack(
        ROUND_LABELS, CH_LABELS, ZPLANE_LABELS, HEIGHT, WIDTH, crop_parameters)


def test_crop_rcz():
    """Build an imagestack that contains a crop in r/c/z.  Verify that the appropriate tiles are
    loaded.
    """
    rounds = [1]
    chs = [2, 3]

    crop_parameters = CropParameters(
        permitted_rounds=rounds,
        permitted_chs=chs,
    )
    stack = setup_imagestack(crop_parameters)

    assert stack.axis_labels(Axes.ROUND) == rounds
    assert stack.axis_labels(Axes.CH) == chs
    assert stack.axis_labels(Axes.ZPLANE) == ZPLANE_LABELS

    expected_zplane_coordinates = np.linspace(Z_COORDS[0], Z_COORDS[1], NUM_ZPLANE)

    for zplane_label in stack.axis_labels(Axes.ZPLANE):
        for round_label in stack.axis_labels(Axes.ROUND):
            for ch_label in stack.axis_labels(Axes.CH):
                expected_tile_data = expected_data(round_label, ch_label, zplane_label)

                verify_stack_data(
                    stack,
                    {Axes.ROUND: round_label, Axes.CH: ch_label, Axes.ZPLANE: zplane_label},
                    expected_tile_data,
                )

        zplane_index = ZPLANE_LABELS.index(zplane_label)
        expected_zplane_coordinate = expected_zplane_coordinates[zplane_index]
        verify_physical_coordinates(
            stack,
            X_COORDS,
            Y_COORDS,
            expected_zplane_coordinate,
            zplane_label,
        )


def test_crop_xy():
    """Build an imagestack that contains a crop in x/y.  Verify that the data is sliced correctly.
    """
    X_SLICE = (10, 30)
    Y_SLICE = (15, 40)
    crop_parameters = CropParameters(
        x_slice=slice(*X_SLICE),
        y_slice=slice(*Y_SLICE),
    )
    stack = setup_imagestack(crop_parameters)

    assert stack.axis_labels(Axes.ROUND) == ROUND_LABELS
    assert stack.axis_labels(Axes.CH) == CH_LABELS
    assert stack.axis_labels(Axes.ZPLANE) == ZPLANE_LABELS

    assert stack.raw_shape[3] == Y_SLICE[1] - Y_SLICE[0]
    assert stack.raw_shape[4] == X_SLICE[1] - X_SLICE[0]

    expected_zplane_coordinates = np.linspace(Z_COORDS[0], Z_COORDS[1], NUM_ZPLANE)

    for zplane in stack.axis_labels(Axes.ZPLANE):
        for round_ in stack.axis_labels(Axes.ROUND):
            for ch in stack.axis_labels(Axes.CH):
                expected_tile_data = expected_data(round_, ch, zplane)
                expected_tile_data = expected_tile_data[
                    Y_SLICE[0]:Y_SLICE[1], X_SLICE[0]:X_SLICE[1]]

                verify_stack_data(
                    stack,
                    {Axes.ROUND: round_, Axes.CH: ch, Axes.ZPLANE: zplane},
                    expected_tile_data,
                )

        # the coordinates should be rescaled.  verify that the coordinates on the ImageStack
        # are also rescaled.
        original_x_coordinates = X_COORDS
        expected_x_coordinates = recalculate_physical_coordinate_range(
            original_x_coordinates[0], original_x_coordinates[1],
            WIDTH,
            slice(*X_SLICE),
        )

        original_y_coordinates = Y_COORDS
        expected_y_coordinates = recalculate_physical_coordinate_range(
            original_y_coordinates[0], original_y_coordinates[1],
            HEIGHT,
            slice(*Y_SLICE),
        )

        zplane_index = ZPLANE_LABELS.index(zplane)
        expected_zplane_coordinate = expected_zplane_coordinates[zplane_index]

        verify_physical_coordinates(
            stack,
            expected_x_coordinates,
            expected_y_coordinates,
            expected_zplane_coordinate,
            zplane,
        )



import random

from starfish import IntensityTable
from starfish.core.codebook.test.factories import codebook_array_factory
from starfish.core.types import Features

NUMBER_SPOTS = 10


def test_save_expression_matrix():

    codebook = codebook_array_factory()

    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=3,
        height=100,
        width=100,
        n_spots=10
    )
    # mock out come cell_ids
    cell_ids = random.sample(range(1, 20), NUMBER_SPOTS)
    intensities[Features.CELL_ID] = (Features.AXIS, cell_ids)

    expression_matrix = intensities.to_expression_matrix()

    # test all saving methods
    expression_matrix.save("expression")



from starfish.core.codebook.codebook import Codebook
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Number
from starfish.core.util import click
from ._base import DecodeAlgorithmBase


# TODO ambrosejcarr add tests
class MetricDistance(DecodeAlgorithmBase):
    """
    Normalizes both the magnitudes of the codes and the spot intensities, then decodes spots by
    assigning each spot to the closest code, measured by the provided metric.

    Codes greater than max_distance from the nearest code, or dimmer than min_intensity, are
    discarded.

    Parameters
    ----------
    codebook : Codebook
        codebook containing targets the experiment was designed to quantify
    max_distance : Number
        spots greater than this distance from their nearest target are not decoded
    min_intensity : Number
        spots dimmer than this intensity are not decoded
    norm_order : int
        the norm to use to normalize the magnitudes of spots and codes (default 2, L2 norm)
    metric : str
        the metric to use to measure distance. Can be any metric that satisfies the triangle
        inequality that is implemented by :py:mod:`scipy.spatial.distance` (default "euclidean")
    """

    def __init__(
        self,
        codebook: Codebook,
        max_distance: Number,
        min_intensity: Number,
        norm_order: int = 2,
        metric: str = "euclidean",
    ) -> None:
        self.codebook = codebook
        self.max_distance = max_distance
        self.min_intensity = min_intensity
        self.norm_order = norm_order
        self.metric = metric

    def run(
        self,
        intensities: IntensityTable,
        *args
    ) -> IntensityTable:
        """Decode spots by selecting the max-valued channel in each sequencing round

        Parameters
        ----------
        intensities : IntensityTable
            IntensityTable to be decoded
        codebook : Codebook
            Contains codes to decode IntensityTable

        Returns
        -------
        IntensityTable :
            IntensityTable decoded and appended with Features.TARGET and Features.QUALITY values.

        """
        return self.codebook.decode_metric(
            intensities,
            max_distance=self.max_distance,
            min_intensity=self.min_intensity,
            norm_order=self.norm_order,
            metric=self.metric,
        )

    @staticmethod
    @click.command("MetricDistance")
    @click.option(
        "--max-distance", type=float,
        help="for a detected spot to be assigned a target, it must be within this distance the "
             "nearest target"
    )
    @click.option("--min-intensity", type=float, help="minimum intensity for a spot to be decoded")
    @click.option(
        "--norm-order", type=int, default=2,
        help="norm to equalize the magnitudes of codes and targets"
    )
    @click.option(
        "--metric", type=str, default="euclidean", help="metric used to calculate distance"
    )
    @click.pass_context
    def _cli(ctx, max_distance, min_intensity, norm_order, metric):
        codebook = ctx.obj["codebook"]
        instance = MetricDistance(
            codebook=codebook,
            max_distance=max_distance,
            min_intensity=min_intensity,
            norm_order=norm_order,
            metric=metric
        )
        ctx.obj["component"]._cli_run(ctx, instance)



from typing import Tuple

import numpy as np
from slicedimage import Tile, TileSet

import starfish.data
from starfish.core.test.factories import SyntheticData
from starfish.types import Axes, Coordinates
from ..experiment import Experiment, FieldOfView


def round_to_x(r: int) -> Tuple[float, float]:
    return (r + 1) * 1000, (r + 1) * 100


def round_to_y(r: int) -> Tuple[float, float]:
    return (r + 1) * 10, (r + 1) * 0.1


def round_to_z(r: int) -> Tuple[float, float]:
    return (r + 1) * 0.01, (r + 1) * 0.001


NUM_ROUND = 5
NUM_CH = 2
NUM_Z = 1
HEIGHT = 100
WIDTH = 100


def get_aligned_tileset():
    alignedTileset = TileSet(
        [Axes.X, Axes.Y, Axes.CH, Axes.ZPLANE, Axes.ROUND],
        {Axes.CH: NUM_CH, Axes.ROUND: NUM_ROUND, Axes.ZPLANE: NUM_Z},
        {Axes.Y: HEIGHT, Axes.X: WIDTH})

    for r in range(NUM_ROUND):
        for ch in range(NUM_CH):
            for z in range(NUM_Z):
                tile = Tile(
                    {
                        Coordinates.X: 1,
                        Coordinates.Y: 4,
                        Coordinates.Z: 3,
                    },
                    {
                        Axes.ROUND: r,
                        Axes.CH: ch,
                        Axes.ZPLANE: z,
                    }
                )
                tile.numpy_array = np.zeros((100, 100))
                alignedTileset.add_tile(tile)
    return alignedTileset


def get_un_aligned_tileset():
    unAlignedTileset = TileSet(
        [Axes.X, Axes.Y, Axes.CH, Axes.ZPLANE, Axes.ROUND],
        {Axes.CH: NUM_CH, Axes.ROUND: NUM_ROUND, Axes.ZPLANE: NUM_Z},
        {Axes.Y: HEIGHT, Axes.X: WIDTH})

    for r in range(NUM_ROUND):
        for ch in range(NUM_CH):
            for z in range(NUM_Z):
                tile = Tile(
                    {
                        # The round_to methods generate coordinates
                        # based on the r value, therefore the coords vary
                        # throughout the tileset
                        Coordinates.X: round_to_x(r),
                        Coordinates.Y: round_to_y(r),
                        Coordinates.Z: round_to_z(r),
                    },
                    {
                        Axes.ROUND: r,
                        Axes.CH: ch,
                        Axes.ZPLANE: z,
                    }
                )
                tile.numpy_array = np.zeros((HEIGHT, WIDTH))
                unAlignedTileset.add_tile(tile)
    return unAlignedTileset


def test_fov_order():
    data = SyntheticData()
    codebook = data.codebook()
    tilesets = {"primary": get_aligned_tileset()}
    fovs = [FieldOfView("stack2", tilesets),
            FieldOfView("stack1", tilesets)]
    extras = {"synthetic": True}
    experiment = Experiment(fovs, codebook, extras)
    assert "stack1" == experiment.fov().name
    assert ["stack1", "stack2"] == [x.name for x in experiment.fovs()]


def test_crop_experiment():
    exp = starfish.data.ISS(use_test_data=True)
    image = exp['fov_001'].get_image('primary', x=slice(10, 30), y=slice(40, 70))
    assert image.shape['x'] == 20
    assert image.shape['y'] == 30

    image = exp['fov_001'].get_image('primary', rounds=[0, 1], chs=[2, 3])
    assert image.num_rounds == 2
    assert image.axis_labels(Axes.ROUND) == [0, 1]
    assert image.num_chs == 2
    assert image.axis_labels(Axes.CH) == [2, 3]


def test_fov_aligned_tileset():
    tilesets = {'primary': get_aligned_tileset(), 'nuclei': get_aligned_tileset()}
    fov = FieldOfView("aligned", tilesets)
    primary_images = fov.get_images('primary')
    nuclei_images = fov.get_images('nuclei')
    # Assert that only one ImageStack in list
    assert len(primary_images) == 1
    assert len(nuclei_images) == 1


def test_fov_un_aligned_tileset():
    tilesets = {'primary': get_un_aligned_tileset(), 'nuclei': get_un_aligned_tileset()}
    fov = FieldOfView("unaligned", tilesets)
    primary_images = fov.get_images('primary')
    nuclei_images = fov.get_images('nuclei')
    # Assert that the number of coordinate groups == NUM_ROUNDS
    assert len(primary_images) == NUM_ROUND
    assert len(nuclei_images) == NUM_ROUND



import warnings
from functools import partial
from typing import Dict, List, NamedTuple, Optional, Tuple

import numpy as np
import pandas as pd
from skimage.measure import label, regionprops
from skimage.measure._regionprops import _RegionProperties
from tqdm import tqdm

from starfish.core.config import StarfishConfig
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.multiprocessing.pool import Pool
from starfish.core.types import Axes, Features, Number, SpotAttributes


class ConnectedComponentDecodingResult(NamedTuple):
    region_properties: List
    label_image: np.ndarray
    decoded_image: np.ndarray


class TargetsMap:

    def __init__(self, targets: np.ndarray) -> None:
        """
        Creates an invertible mapping between string names of Codebook targets and integer IDs
        that can be interpreted by skimage.measure to decode an image.

        Parameters
        ----------
        targets : np.ndarray
            array of string target IDs

        """
        unique_targets = set(targets) - {'nan'}
        sorted_targets = sorted(unique_targets)
        self._int_to_target = dict(zip(range(1, np.iinfo(np.int).max), sorted_targets))
        self._int_to_target[0] = 'nan'
        self._target_to_int = {v: k for (k, v) in self._int_to_target.items()}

    def targets_as_int(self, targets: np.ndarray) -> np.ndarray:
        """Transform an array of targets into their integer representation.

        Parameters
        ----------
        targets : np.ndarray['U']
            array of string targets to be transformed into integer IDs

        Returns
        -------
        np.ndarray[int] :
            array of targets represented by their integer IDs

        """
        return np.array([self._target_to_int[v] for v in targets])

    def targets_as_str(self, targets: np.ndarray) -> np.ndarray:
        """Transform an array of integer IDs into their corresponding string target names.

        Parameters
        ----------
        targets : np.ndarray[int]
            array of int targets to be transformed into string names

        Returns
        -------
        np.ndarray['U']
            array of unicode-encoded target names

        """
        return np.array([self._int_to_target[v] for v in targets])

    def target_as_str(self, integer_target: int) -> np.ndarray:
        return self._int_to_target[integer_target]


class CombineAdjacentFeatures:

    def __init__(
            self,
            min_area: Number,
            max_area: Number,
            connectivity: int=2,
            mask_filtered_features: bool=True
    ) -> None:
        """Combines pixel-wise adjacent features into single larger features using skimage.measure

        Parameters
        ----------
        min_area : Number
            Combined features with area below this value are marked as failing filters
        max_area : Number
            Combined features with area above this value are marked as failing filters
        connectivity : int
            Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor. See
            http://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.label for more
            detail. Default = 2.
        mask_filtered_features : bool
            If True, sets all pixels that are failing filters applied prior to this function equal
            to zero, the background value for skimage.measure.label

        """
        self._min_area = min_area
        self._max_area = max_area
        self._connectivity = connectivity
        self._mask_filtered = mask_filtered_features

    @staticmethod
    def _intensities_to_decoded_image(
            intensities: IntensityTable,
            target_map: TargetsMap,
            mask_filtered_features: bool=True
    ) -> np.ndarray:
        """
        Construct an image where each pixel corresponds to its decoded target, mapped to a unique
        integer ID

        Parameters
        ----------
        intensities : IntensityTable
            Decoded intensities
        target_map : TargetsMap
            Mapping between string target names and integer target IDs
        mask_filtered_features : bool
            If true, all features that fail filters are mapped to zero, which is considered
            'background' and will not decode to a target (default = True).

        Returns
        -------
        np.ndarray[int]
            Image whose pixels are coded as the targets that the ImageStack decoded to at each
            position.

        """
        # reverses the linearization that was used to transform an ImageStack into an IntensityTable
        max_x = intensities[Axes.X.value].values.max() + 1
        max_y = intensities[Axes.Y.value].values.max() + 1
        max_z = intensities[Axes.ZPLANE.value].values.max() + 1

        int_targets = target_map.targets_as_int(intensities[Features.TARGET].values)
        if mask_filtered_features:
            fails_filters = np.where(~intensities[Features.PASSES_THRESHOLDS])[0]
            int_targets[fails_filters] = 0

        decoded_image: np.ndarray = int_targets.reshape((max_z, max_y, max_x))
        return decoded_image

    @staticmethod
    def _calculate_mean_pixel_traces(
            label_image: np.ndarray,
            intensities: IntensityTable,
    ) -> IntensityTable:
        """
        For all pixels that contribute to a connected component, calculate the mean value for
        each (ch, round), producing an average "trace" of a feature across the imaging experiment

        Parameters
        ----------
        label_image : np.ndarray
            An image where all pixels of a connected component share the same integer ID
        intensities : IntensityTable
            decoded intensities

        Returns
        -------
        IntensityTable :
            an IntensityTable where the number of features equals the number of connected components
            and the intensities of each each feature is its mean trace.

        """

        import xarray as xr
        pixel_labels = label_image.reshape(-1)

        # Use a pandas groupby approach-based approach, because it is much faster than xarray

        # If needed, it is possible to be even faster than pandas:
        # https://stackoverflow.com/questions/51975512/\
        # faster-alternative-to-perform-pandas-groupby-operation

        # stack intensities
        stacked = intensities.stack(traces=(Axes.CH.value, Axes.ROUND.value))

        # drop into pandas to use their faster groupby
        traces: pd.DataFrame = pd.DataFrame(
            stacked.values,
            index=pixel_labels,
            columns=stacked.traces.to_index()
        )

        #
        distances: pd.Series = pd.Series(
            stacked[Features.DISTANCE].values, index=pixel_labels
        )

        grouped = traces.groupby(level=0)
        pd_mean_pixel_traces = grouped.mean()

        grouped = distances.groupby(level=0)
        pd_mean_distances = grouped.mean()

        pd_xarray = xr.DataArray(
            pd_mean_pixel_traces,
            dims=(Features.AXIS, 'traces'),
            coords=dict(
                traces=('traces', pd_mean_pixel_traces.columns),
                distance=(Features.AXIS, pd_mean_distances),
                features=(Features.AXIS, pd_mean_pixel_traces.index)
            )
        )
        mean_pixel_traces = pd_xarray.unstack('traces')

        # the 0th pixel trace corresponds to background. If present, drop it.
        try:
            mean_pixel_traces = mean_pixel_traces.drop(0, dim=Features.AXIS)
        except KeyError:
            pass

        return mean_pixel_traces

    @staticmethod
    def _single_spot_attributes(
            spot_property: _RegionProperties,
            decoded_image: np.ndarray,
            target_map: TargetsMap,
            min_area: Number,
            max_area: Number,
    ) -> Tuple[Dict[str, int], int]:
        """
        Calculate starfish SpotAttributes from the RegionProperties of a connected component
        feature.

        Parameters
        ----------
        spot_property: _RegionProperties
            Properties of the connected component. Output of skimage.measure.regionprops
        decoded_image : np.ndarray
            Image whose pixels correspond to the targets that the given position in the ImageStack
            decodes to.
        target_map : TargetsMap
            Unique mapping between string target names and int target IDs.
        min_area :
            Combined features with area below this value are marked as failing filters
        max_area : Number
            Combined features with area above this value are marked as failing filters

        Returns
        -------
        Dict[str, Number] :
            spot attribute dictionary for this connected component, containing the x, y, z position,
            target name (str) and feature radius.
        int :
            1 if spot passes size filters, zero otherwise.

        """
        # because of the above skimage issue, we need to support both 2d and 3d properties
        centroid = spot_property.centroid
        if len(centroid) == 3:
            spot_attrs = {
                'z': int(centroid[0]),
                'y': int(centroid[1]),
                'x': int(centroid[2])
            }
        else:  # data is 2d
            spot_attrs = {
                'z': 0,
                'y': int(centroid[0]),
                'x': int(centroid[1])
            }

        # define the target index as the most repeated value in the bounding box of the spot.
        #  it appears there is no simpler way to do this with a regionprops object
        bbox = spot_property.bbox
        if len(bbox) == 6:
            # 3d bbox
            target_candidates = np.ndarray.flatten(
                decoded_image[bbox[0]:bbox[3], bbox[1]:bbox[4], bbox[2]:bbox[5]])
        else:
            # flatten and remove zeros
            target_candidates = np.ndarray.flatten(
                decoded_image[0, bbox[0]:bbox[2], bbox[1]:bbox[3]])
        # get the most repeated nonzero value
        non_zero_target_candidates = target_candidates[target_candidates != 0]
        target_index = np.argmax(np.bincount(non_zero_target_candidates))
        spot_attrs[Features.TARGET] = target_map.target_as_str(target_index)
        spot_attrs[Features.SPOT_RADIUS] = spot_property.equivalent_diameter / 2

        # filter intensities for which radius is too small
        passes_area_filter = 1 if min_area <= spot_property.area < max_area else 0
        return spot_attrs, passes_area_filter

    def _create_spot_attributes(
            self,
            region_properties: List[_RegionProperties],
            decoded_image: np.ndarray,
            target_map: TargetsMap,
            n_processes: Optional[int]=None
    ) -> Tuple[SpotAttributes, np.ndarray]:
        """

        Parameters
        ----------
        region_properties : List[_RegionProperties]
            Properties of the each connected component. Output of skimage.measure.regionprops
        decoded_image : np.ndarray
            Image whose pixels correspond to the targets that the given position in the ImageStack
            decodes to.
        target_map : TargetsMap
            Unique mapping between string target names and int target IDs.
        n_processes : Optional[int]=None
            number of processes to devote to measuring spot properties. If None, defaults to the
            result of os.nproc()

        Returns
        -------
        pd.DataFrame :
            DataFrame containing x, y, z, radius, and target name for each connected component
            feature.
        np.ndarray[bool] :
            An array with length equal to the number of features. If zero, indicates that a feature
            has failed area filters.
        """
        with Pool(processes=n_processes) as pool:
            mapfunc = pool.map
            applyfunc = partial(
                self._single_spot_attributes,
                decoded_image=decoded_image,
                target_map=target_map,
                min_area=self._min_area,
                max_area=self._max_area
            )

            iterable = tqdm(region_properties, disable=(not StarfishConfig().verbose))
            results = mapfunc(applyfunc, iterable)
            if not results:
                # no spots found
                warnings.warn("No spots found, please adjust threshold parameters")
                return SpotAttributes.empty(extra_fields=['target']), np.array(0, dtype=np.bool)
            spot_attrs, passes_area_filter = zip(*results)

            # update passes filter
            passes_filter = np.array(passes_area_filter, dtype=np.bool)

            spot_attributes = SpotAttributes(pd.DataFrame.from_records(spot_attrs))
            return spot_attributes, passes_filter

    def run(
            self, intensities: IntensityTable,
            n_processes: Optional[int] = None,
    ) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """
        Execute the combine_adjacent_features method on an IntensityTable containing pixel
        intensities

        Parameters
        ----------
        intensities : IntensityTable
            Pixel intensities of an imaging experiment
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        IntensityTable :
            Table whose features comprise sets of adjacent pixels that decoded to the same target
        ConnectedComponentDecodingResult :
            NamedTuple containing :
                region_properties :
                    the properties of each connected component, in the same order as the
                    IntensityTable
                label_image : np.ndarray
                    An image where all pixels of a connected component share the same integer ID
                decoded_image : np.ndarray
                    Image whose pixels correspond to the targets that the given position in the
                    ImageStack decodes to.

        """

        # map target molecules to integers so they can be reshaped into an image that can
        # be subjected to a connected-component algorithm to find adjacent pixels with the
        # same targets
        targets = intensities[Features.TARGET].values
        target_map = TargetsMap(targets)

        # create the decoded_image
        decoded_image = self._intensities_to_decoded_image(
            intensities,
            target_map,
            self._mask_filtered,
        )

        # label the decoded image to extract connected component features
        label_image: np.ndarray = label(decoded_image, connectivity=self._connectivity)

        # calculate properties of each feature
        props: List = regionprops(np.squeeze(label_image))

        # calculate mean intensities across the pixels of each feature
        mean_pixel_traces = self._calculate_mean_pixel_traces(
            label_image,
            intensities,
        )

        # Create SpotAttributes and determine feature filtering outcomes
        spot_attributes, passes_filter = self._create_spot_attributes(
            props,
            decoded_image,
            target_map,
            n_processes=n_processes
        )

        # augment the SpotAttributes with filtering results and distances from nearest codes
        spot_attributes.data[Features.DISTANCE] = mean_pixel_traces[Features.DISTANCE]
        spot_attributes.data[Features.PASSES_THRESHOLDS] = passes_filter

        # create new indexes for the output IntensityTable
        channel_index = mean_pixel_traces.indexes[Axes.CH]
        round_index = mean_pixel_traces.indexes[Axes.ROUND]
        coords = IntensityTable._build_xarray_coords(spot_attributes, channel_index, round_index)

        # create the output IntensityTable
        dims = (Features.AXIS, Axes.CH.value, Axes.ROUND.value)
        intensity_table = IntensityTable(
            data=mean_pixel_traces, coords=coords, dims=dims
        )

        # combine the various non-IntensityTable results into a NamedTuple before returning
        ccdr = ConnectedComponentDecodingResult(props, label_image, decoded_image)

        return intensity_table, ccdr
